from __future__ import absolute_import
import os
import ee
import math
import copy
import datetime
from pipetools import pipe
from pprint import pformat
from functools import partial
from ee.ee_exception import EEException
from hydrafloods import decorators, geeutils


class Dataset:
    """Base dataset class used to define an EE image collection by datetime and geographic region

    A dataset wraps an ee.ImageCollection by applying the spatial and temporal filtering upon init.
    Provides utility functionality to make working with and managing image collections less verbose

    Example:
        Create a dataset object for Sentinel-1 data over Alabama for 2019
        >>> ds = Dataset(
        ...    region = ee.Geometry.Rectangle([
                   -88.473227,
                   30.223334,
                   -84.88908,
                   35.008028
               ]),
        ...    start_time = "2019-01-01",
        ...    end_time = "2020-01-01",
        ...    asset_id = "COPERNICUS/S1_GRD"
        ... )
        >>> ds
        HYDRAFloods Dataset:
        {'asset_id': 'COPERNICUS/S1_GRD',
        'end_time': '2020-01-01',
        'name': 'Dataset',
        'region': [[[...], [...], [...], [...], [...]]],
        'start_time': '2019-01-01'}
    """

    def __init__(self, region, start_time, end_time, asset_id, use_qa=False):
        """Initialize Dataset class

        args:
            region (ee.Geometry): earth engine geometry object to filter image collection by
            start_time (str | datetime.datetime): start time used to filter image collection
            end_time (str | datetime.datetime): end time used to filter image collection
            asset_id (str): asset id of earth engine collection
            use_qa (bool, optional): boolean to determine to use an internal function qa().
                Used for definining custom dataset objects

        raises:
            AttributeError: if qa() method is not defined and use_qa is True
        """

        # TODO: add exceptions to check datatypes
        self.region = region  # dtype = ee.Geometry
        self.start_time = start_time
        self.end_time = end_time
        self.asset_id = asset_id
        self.use_qa = use_qa

        # dictionary mapping of band names used to harmonized optical datasets to same names
        self.BANDREMAP = ee.Dictionary(
            {
                "landsat7": ee.List(["SR_B1", "SR_B2", "SR_B3", "SR_B4", "SR_B5", "SR_B7"]),
                "landsat8": ee.List(["SR_B2", "SR_B3", "SR_B4", "SR_B5", "SR_B6", "SR_B7"]),
                "viirs": ee.List(["M2", "M4", "I1", "I2", "I3", "M11"]),
                "sen2": ee.List(["B2", "B3", "B4", "B8", "B11", "B12"]),
                "modis": ee.List(
                    [
                        "sur_refl_b03",
                        "sur_refl_b04",
                        "sur_refl_b01",
                        "sur_refl_b02",
                        "sur_refl_b06",
                        "sur_refl_b07",
                    ]
                ),
                "new": ee.List(["blue", "green", "red", "nir", "swir1", "swir2"]),
            }
        )

        # get the image collection and filter by geographic region and date time
        imgcollection = (
            ee.ImageCollection(self.asset_id)
            .filterBounds(self.region)
            .filterDate(self.start_time, self.end_time)
        )

        # check if to apply arbitrary qa process on collection
        # qa function can be defined in custom objects extending dataset
        if self.use_qa:
            try:
                imgcollection = imgcollection.map(self.qa)
            except AttributeError:
                raise AttributeError(
                    "qa() method is not defined...please define one or set `use_qa` to False"
                )

        self.collection = imgcollection

    def __repr__(self):
        # format datetime information
        if isinstance(self.start_time, datetime.datetime):
            ststr = self.start_time.strftime("%Y-%m-%d")
        else:
            ststr = self.start_time

        if isinstance(self.end_time, datetime.datetime):
            etstr = self.end_time.strftime("%Y-%m-%d")
        else:
            etstr = self.end_time

        # create dict of dataset information
        objDict = {
            "name": self.__class__.__name__,
            "asset_id": self.asset_id,
            "start_time": ststr,
            "end_time": etstr,
            "region": self.region.coordinates().getInfo(),
        }

        # pretty format dict and return information
        strRepr = pformat(objDict, depth=3)
        return f"HYDRAFloods Dataset:\n{strRepr}"

    @property
    def collection(self):
        """image collection object property wrapped by dataset"""
        return self._collection

    @collection.setter
    def collection(self, value):
        """setter function for collection property"""
        self._collection = value
        return

    @property
    def n_images(self):
        """Number of images contained in the dataset"""
        return self.collection.size().getInfo()

    @property
    def dates(self):
        """Dates of imagery contained in the image collection"""
        eeDates = self.collection.aggregate_array("system:time_start").map(
            lambda x: ee.Date(x).format("YYYY-MM-dd HH:mm:ss.SSS")
        )
        return eeDates.getInfo()

    @staticmethod
    def from_imgcollection(img_collection):
        """Static method to convert an ee.ImageCollection object to a hf.Dataset object.
        This method will take some time as it uses computed ee Objects from the image collection
        propeties to populate the Dataset object properties (passing info from server to client)

        args:
            img_collection (ee.ImageCollection): computed ee.ImageCollection object to create a hf.Dataset

        returns:
            hf.Dataset: dataset object with property information directly from the ee.ImageCollection
        """

        # get region and date information
        region = (
            img_collection.map(geeutils.get_geoms)
            .union(maxError=100)
            .geometry(maxError=100)
        )
        # convert ee.Date info to string format
        dates = (
            img_collection.aggregate_array("system:time_start")
            .sort()
            .map(lambda x: ee.Date(x).format("YYYY-MM-dd HH:mm:ss.S"))
        )

        # pull the date info to local strings
        start_time = dates.get(0).getInfo()
        end_time = dates.get(-1).getInfo()

        # get the collection id for `.asset_id` property
        collection_id = img_collection.get("system:id").getInfo()
        if collection_id == "None":
            collection_id = "Custom ImageCollection"

        # make a dummy dataset
        dummy_ds = Dataset(
            region,
            start_time,
            end_time,
            asset_id="NOAA/VIIRS/001/VNP09GA",
            use_qa=False,
        )

        # override the dummy dataset information with the correct data fr
        dummy_ds.asset_id = collection_id
        dummy_ds.collection = img_collection

        return dummy_ds

    def qa(
        self,
    ):
        return

    def _inplace_wrapper(self, collection, inplace):
        """Private helper function to replace the collection info for a class
        typically used when other class methods have inplace
        """
        if inplace:
            self.collection = collection
            return
        else:
            outCls = self.copy()
            outCls.collection = collection
            return outCls

    def copy(self):
        """returns a deep copy of the hydrafloods dataset class"""
        return copy.deepcopy(self)

    def apply_func(self, func, inplace=False, *args, **kwargs):
        """Wrapper method to apply a function to all of the image in the dataset.
        Makes a copy of the collection and reassigns the image collection property.
        Function must accept an ee.ImageCollection and return an ee.ImageCollection

        args:
            func (object): Function to map across image collection. Function must accept ee.Image as first argument
            inplace (bool, optional): define whether to return another dataset object or update inplace. default = False
            **kwargs: arbitrary keyword to pass to `func`

        returns:
            Dataset | None: copy of class with results from `func` as image within collection property
        """

        # get a partial function to map over imagery with the keywords applied
        # expects that the first positional arg is an
        func = partial(func, **kwargs)

        out_coll = self.collection.map(func)

        return self._inplace_wrapper(out_coll, inplace)

    def apply(self, func, inplace=False, *args, **kwargs):
        """Alias for the `apply_func` method

        args:
            func (object): Function to map across image collection. Function must accept ee.Image as first argument
            inplace (bool, optional): define whether to return another dataset object or update inplace. default = False
            **kwargs: arbitrary keyword to pass to `func`

        returns:
            Dataset | None: copy of class with results from `func` as image within collection property
        """

        return self.apply_func(func, inplace, *args, *kwargs)

    def filter(self, filter, inplace=False):
        """Wrapper method for applying a filter to a datset collection.

        args:
            filter (ee.Filter): an `ee.Filter` object to apply to the dataset collection.

        returns:
            Dataset | None: returns the dataset with the filtered collection.
        """

        filtered = self.collection.filter(filter)

        return self._inplace_wrapper(filtered, inplace)

    def select(self, *args, inplace=False):
        """Wrapper method for selecting bands from dataset collection

        args:
            *args: arbitrary arguments to pass to the `ee.ImageCollection.select()` method
            inplace (bool, optional): define whether to return another dataset object or update inplace. default = False

        returns:
            Dataset | None: returns the dataset with the collection with selected bands.

        """
        selected = self.collection.select(*args)

        return self._inplace_wrapper(selected, inplace)

    def clip_to_region(self, inplace=False):
        """Clips all of the images to the geographic extent defined by region.
        Useful for setting geometries on unbounded imagery in collection (e.g. MODIS or VIIRS imagery)

        args:
            inplace (bool, optional): define whether to return another dataset object or update inplace. default = False

        returns:
            Dataset | None: returns dataset with imagery clipped to self.region or none depending on inplace
        """

        @decorators.keep_attrs
        def clip(img):
            """Closure function to perform the clipping while carrying metadata"""
            return ee.Image(img.clip(self.region))

        out_coll = self.collection.map(clip)

        return self._inplace_wrapper(out_coll, inplace)

    def merge(self, dataset, inplace=False):
        """Merge the collection of two datasets into one where self.collection will contain imagery from self and dataset arg.
        Results will be sorted by time

        args:
            dataset (Dataset): dataset object to merge
            inplace (bool, optional): define whether to return another dataset object or update inplace. default = False

        returns:
            Dataset | None: returns dataset where collection is merged imagery or none depending on inplace

        """
        merged = self.collection.merge(dataset.collection).sort("system:time_start")

        return self._inplace_wrapper(merged, inplace)

    def join(self, dataset, inplace=False):
        """Performs spatiotemporal join between self.collection and dataset.collection.
        Result will be a dataset where the collection is colocated imagery in space and time

        args:
            dataset (Dataset): dataset object to apply join with. Used as right in join operations
            inplace (bool, optional): define whether to return another dataset object or update inplace. default = False

        returns:
            Dataset | None: returns dataset where collection with joined imagery or none depending on inplace

        """

        def _merge(img):
            """Closure func to take results from the join and combine into one image with overlaping region"""
            join_coll = ee.ImageCollection.fromImages(img.get(key))

            img_geom = img.geometry(100)
            join_geom = join_coll.map(geeutils.get_geoms).union(100).geometry(100)
            overlap = img_geom.intersection(join_geom, 100)

            join_data = join_coll.mosaic()

            return img.addBands(join_data).clip(overlap)

        key = str(dataset.__class__.__name__)

        # get a time and space filter
        filter = ee.Filter.And(
            ee.Filter.maxDifference(
                **{
                    "difference": 1000 * 60 * 60 * 24,  # One day in milliseconds
                    "leftField": "system:time_start",
                    "rightField": "system:time_start",
                }
            ),
            ee.Filter.intersects(
                **{"leftField": ".geo", "rightField": ".geo", "maxError": 100}
            ),
        )
        # apply join on collections and save all results
        joined = ee.ImageCollection(
            ee.Join.saveAll(key).apply(
                primary=self.collection, secondary=dataset.collection, condition=filter
            )
        )

        # map over all filtered imagery, mosaic joined matches, and add bands to imagery
        joined = joined.map(_merge)

        return self._inplace_wrapper(joined, inplace)

    def aggregate_time(
        self,
        dates=None,
        period=1,
        period_unit="day",
        reducer="mean",
        rename=True,
        clip_to_area=False,
        inplace=False,
    ):
        """Aggregates multiple images into one based on time periods and a user defined reducer.
        Useful for mosaicing images from same date or time period.
        Expects the images in this dataset to be homogenous (same band names and types)

        args:
            dates (list[str], optional): list of dates defined as beginning time period of aggregatation. default = None,
                all available uniques dates in collection will be used
            period (int, optional): number of days to advance from dates for aggregation. default = 1
            period_unit (str, optional): time unit to advance period for aggregation. default = "day"
            reducer (str | ee.Reducer, optional): reducer to apply to images for aggregation, accepts string reducer name
                or ee.Reducer opbject, default = "mean"
            clip_to_area (bool): switch to clip imagery that has been merged to the overlaping region of imagery, default=False
            inplace (bool, optional): define whether to return another dataset object or update inplace. default = False

        returns:
            Dataset | None: returns dataset.collection with aggregated imagery or none depending on inplace

        """

        def _aggregation(d):
            """Closure function to map through days and reduce data within a given time period"""
            t1 = ee.Date(d)
            t2 = t1.advance(period, period_unit)
            img = (
                self.collection.filterDate(t1, t2)
                .reduce(reducer)
                .set("system:time_start", t1.millis())
            )
            if rename:
                img = img.rename(band_names)

            geom = (
                ee.FeatureCollection(
                    self.collection.filterDate(t1, t2).map(geeutils.get_geoms)
                )
                .union(100)
                .geometry(100)
            )
            outimg = ee.Algorithms.If(clip_to_area, img.clip(geom), img)

            return outimg

        if dates is None:
            dates = (
                self.collection.aggregate_array("system:time_start")
                .map(lambda x: ee.Date(x).format("YYYY-MM-dd"))
                .distinct()
            )
        else:
            dates = ee.List(dates)

        band_names = ee.Image(self.collection.first()).bandNames()

        out_coll = ee.ImageCollection.fromImages(dates.map(_aggregation))

        return self._inplace_wrapper(out_coll, inplace)

    @decorators.keep_attrs
    def band_pass_adjustment(self, img):
        """Method to apply linear band transformation to dataset image collection.
        Expects that dataset has properties `self.gain` and `self.bias` set

        args:
            img (ee.Image): image to apply regression on

        """
        # linear regression coefficients for adjustment
        return (
            img.multiply(self.gain)
            .add(self.bias)
            .set("system:time_start", img.get("system:time_start"))
        )

    def pipe(self, steps, inplace=False, keep_attrs=True):
        """Method to pipe imagery within dataset through multiple functions at once.
        Assumes the first argument into piped functions are and ee.Image

        args:
            steps (list | tuple): iterable of functions/steps to apply to imagery.
                list must be in the form of (func,func) or with a tuple of function/keyword ((func,kwargs),func)
            inplace (bool, optional): define whether to return another dataset object or update inplace. default = False

        returns:
            Dataset | None: returns dataset.collection with piped functions applied

        Example:
            ```python
            s1 = hf.Sentinel1(ee.Geometry.Point(105.03,11.72),"2019-10-03","2019-10-05")
            water = s1.pipe(
                (
                    hf.gamma_map, #apply speckle filter
                    (hf.egde_ostu,{'initial_threshold:-16}) # apply water mapping
                )
            )
            ```
        """

        def _piper(funcs):
            """Closure function to nest list of functions"""
            if len(funcs) > 1:
                one_shotter = funcs[0]
                for func in funcs[1:]:
                    one_shotter = pipe | one_shotter | func

            else:
                one_shotter = funcs[0]

            return one_shotter

        fs = []
        # loop through the steps and create partial funcs is kwargs are provided
        for step in steps:
            try:
                func, kwargs = step

            except TypeError:
                func = step
                kwargs = None

            if kwargs is not None:
                pfunc = partial(func, **kwargs)
            else:
                pfunc = func

            fs.append(pfunc)

        # get the piped function
        if keep_attrs:
            one_shot = decorators.keep_attrs(_piper(fs))
        else:
            one_shot = _piper(fs)

        # apply pipe to each image
        out_coll = self.collection.map(lambda img: one_shot(img))

        return self._inplace_wrapper(out_coll, inplace)


class Sentinel1(Dataset):
    """Class extending dataset for the Sentinel 1 collection
    This Sentinel 1 dataset is in backscatter units
    """

    def __init__(self, *args, asset_id="COPERNICUS/S1_GRD", use_qa=True, **kwargs):
        """Initialize Sentinel1 Dataset class

        args:
            *args: positional arguments to pass to `Dataset` (i.e. `region`, `start_time`, `end_time`)
            asset_id (str): asset id of the Sentinel 1 earth engine collection. default="COPERNICUS/S1_GRD"
            use_qa (bool, optional): boolean to determine to use a private `self.qa()` function. default=True
            **kwargs (optional): addtional arbitrary keywords to pass to `Dataset`
        """
        super(Sentinel1, self).__init__(
            *args, asset_id=asset_id, use_qa=use_qa, **kwargs
        )

        self.collection = self.collection.filter(
            ee.Filter.listContains("transmitterReceiverPolarisation", "VH")
        )

        return

    @decorators.keep_attrs
    def qa(self, img):
        """Custom QA masking method for Sentinel1 backscatter based on view angle
        Angle threshold values taken from https://doi.org/10.3390/rs13101954
        """
        angle = img.select("angle")
        angle_mask = angle.lt(45.23993).And(angle.gt(30.63993))
        return img.updateMask(angle_mask)

    def add_orbit_band(self, inplace=False):
        """Method to add orbit band from S1 image metadata
        Useful for determining if pixels are from ascending or descending orbits

        args:
            inplace (bool, optional): define whether to return another dataset object or update inplace. default = False

        returns:
            Dataset | None: returns dataset.collection where imagery has the added bands
        """

        def _add_features(img):
            """Closure function to add features as bands to the images"""
            bounds = img.geometry(100)
            orbit = ee.String(img.get("orbitProperties_pass"))
            orbit_band = ee.Algorithms.If(
                orbit.compareTo("DESCENDING"), ee.Image(1), ee.Image(0)
            )

            extraFeatures = ee.Image(orbit_band).rename("orbit")

            return img.addBands(extraFeatures.clip(bounds))

        return self.apply_func(_add_features, inplace=inplace)

    def to_db(self, inplace=False):
        """Convience method to convert units from power to db"""
        out_coll = self.collection.map(geeutils.power_to_db)

        if inplace:
            self.collection = out_coll
            return
        else:
            outCls = self.copy()
            outCls.collection = out_coll
            return outCls

    def to_power(self, inplace=False):
        """Convience method to convert units from db to power"""
        out_coll = self.collection.map(geeutils.db_to_power)

        if inplace:
            self.collection = out_coll
            return
        else:
            outCls = self.copy()
            outCls.collection = out_coll
            return outCls


class Sentinel1Asc(Sentinel1):
    """Class extending dataset for the Sentinel 1 collection for ascending orbit
    This Sentinel 1 dataset is in power units
    """

    def __init__(self, *args, asset_id="COPERNICUS/S1_GRD", use_qa=True, **kwargs):
        """Initialize Sentinel1 Dataset class

        args:
            *args: positional arguments to pass to `Dataset` (i.e. `region`, `start_time`, `end_time`)
            asset_id (str): asset id of the Sentinel 1 earth engine collection. default="COPERNICUS/S1_GRD"
            use_qa (bool, optional): boolean to determine to use a private `self.qa()` function. default=True
            **kwargs (optional): addtional arbitrary keywords to pass to `Dataset`
        """
        super(Sentinel1Asc, self).__init__(
            *args, asset_id=asset_id, use_qa=use_qa, **kwargs
        )

        self.filter(ee.Filter.eq("orbitProperties_pass", "ASCENDING"), inplace=True)

        return


class Sentinel1Desc(Sentinel1):
    """Class extending dataset for the Sentinel 1 collection for descending orbit
    This Sentinel dataset is in power units
    """

    def __init__(self, *args, asset_id="COPERNICUS/S1_GRD", use_qa=True, **kwargs):
        """Initialize Sentinel1 Dataset class

        args:
            *args: positional arguments to pass to `Dataset` (i.e. `region`, `start_time`, `end_time`)
            asset_id (str): asset id of the Sentinel 1 earth engine collection. default="COPERNICUS/S1_GRD"
            use_qa (bool, optional): boolean to determine to use a private `self.qa()` function. default=True
            **kwargs (optional): addtional arbitrary keywords to pass to `Dataset`
        """
        super(Sentinel1Desc, self).__init__(
            *args, asset_id=asset_id, use_qa=use_qa, **kwargs
        )

        self.filter(ee.Filter.eq("orbitProperties_pass", "DESCENDING"), inplace=True)

        return


class Viirs(Dataset):
    def __init__(
        self,
        *args,
        asset_id="NOAA/VIIRS/001/VNP09GA",
        use_qa=True,
        apply_band_adjustment=False,
        **kwargs,
    ):
        """Initialize VIIRS Dataset class

        args:
            *args: positional arguments to pass to `Dataset` (i.e. `region`, `start_time`, `end_time`)
            asset_id (str): asset id of the VIIRS earth engine collection. default="NOAA/VIIRS/001/VNP09GA"
            use_qa (bool, optional): boolean to determine to use a private `self.qa()` function. default=True
            apply_band_adjustment (bool, optional): boolean switch to apply linear band pass equation to convert values to Landsat8. default=False
            rescale (bool, optional): boolean switch to convert units from scaled int (0-10000) to float (0-1). If false values will be scaled int. default = False
            **kwargs (optional): addtional arbitrary keywords to pass to `Dataset`
        """
        super(Viirs, self).__init__(*args, asset_id=asset_id, use_qa=use_qa, **kwargs)

        # get the bands and rename to common optical names
        coll = self.collection.select(
            self.BANDREMAP.get("viirs"), self.BANDREMAP.get("new")
        )

        if apply_band_adjustment:
            # band bass adjustment coefficients taken calculated from https://code.earthengine.google.com/876f53861690e483fb3e3439a3571f27
            # slope coefficients
            self.gain = ee.Image.constant(
                [0.68328, 0.66604, 0.78901, 0.95324, 0.98593, 0.88941]
            )
            # y-intercept coefficients
            self.bias = ee.Image.constant(
                [0.016728, 0.030814, 0.023199, 0.036571, 0.026923, 0.021615]
            ).multiply(10000)
            coll = coll.map(self.band_pass_adjustment)

        self.collection = coll

        return

    @decorators.keep_attrs
    def qa(self, img):
        """Custom QA masking method for VIIRS VNP09GA dataset"""
        cloudMask = geeutils.extract_bits(
            img.select("QF1"), 2, end=3, new_name="cloud_qa"
        ).lt(1)
        shadowMask = geeutils.extract_bits(
            img.select("QF2"), 3, new_name="shadow_qa"
        ).Not()
        snowMask = geeutils.extract_bits(img.select("QF2"), 5, new_name="snow_qa").Not()
        sensorZenith = img.select("SensorZenith").abs().lt(6000)

        env_mask = cloudMask.And(shadowMask).And(sensorZenith)

        # internal pixel quality masks
        pixal_quality_3 = img.select("QF3")
        pixal_quality_4 = img.select("QF4")
        m2_qual = geeutils.extract_bits(pixal_quality_3, 1, new_name="m2_quality").eq(0)
        m4_qual = geeutils.extract_bits(pixal_quality_3, 3, new_name="m4_quality").eq(0)
        m11_qual = geeutils.extract_bits(pixal_quality_4, 0, new_name="m11_quality").eq(0)
        i1_qual = geeutils.extract_bits(pixal_quality_4, 1, new_name="i1_quality").eq(0)
        i2_qual = geeutils.extract_bits(pixal_quality_4, 2, new_name="i2_quality").eq(0)
        i3_qual = geeutils.extract_bits(pixal_quality_4, 3, new_name="i3_quality").eq(0)

        qual_mask = m2_qual.And(m4_qual).And(m11_qual).And(i1_qual).And(i2_qual).And(i3_qual)

        mask = env_mask.And(qual_mask)

        return geeutils.rescale(img).updateMask(mask)


class Modis(Dataset):
    def __init__(
        self, *args, asset_id="MODIS/006/MOD09GA", use_qa=True, **kwargs
    ):
        """Initialize MODIS Dataset class
        Can be used with MOD09GA and MYD09GA

        args:
            *args: positional arguments to pass to `Dataset` (i.e. `region`, `start_time`, `end_time`)
            asset_id (str): asset id of the MODIS earth engine collection. default="MODIS/006/MOD09GA"
            use_qa (bool, optional): boolean to determine to use a private `self.qa()` function. default=True
            rescale (bool, optional): boolean switch to convert units from scaled int (0-10000) to float (0-1). If false values will be scaled int. default = False
            **kwargs (optional): addtional arbitrary keywords to pass to `Dataset`
        """
        super(Modis, self).__init__(*args, asset_id=asset_id, use_qa=use_qa, **kwargs)

        self.collection = self.collection.select(
            self.BANDREMAP.get("modis"), self.BANDREMAP.get("new")
        )

        return

    @decorators.keep_attrs
    def qa(self, img):
        """Custom QA masking method for MODIS MXD09GA dataset"""
        # internal env masks
        qa = img.select("state_1km")
        cloudMask = geeutils.extract_bits(qa, 10, end=11, new_name="cloud_qa").lt(1)
        shadowMask = geeutils.extract_bits(qa, 2, new_name="shadow_qa").Not()
        snowMask = geeutils.extract_bits(qa, 12, new_name="snow_qa").Not()
        cloudAdjMask = geeutils.extract_bits(qa, 13, new_name="cloud_adjacency_qa").Not()
        sensorZenith = img.select("SensorZenith").abs().lt(6000)
        env_mask = cloudMask.And(shadowMask).And(snowMask).And(sensorZenith).And(cloudAdjMask)

        # internal pixel quality masks
        pixal_quality = img.select("QC_500m")
        b1_qual = geeutils.extract_bits(pixal_quality, 2, end=5, new_name="b1_quality").eq(0)
        b2_qual = geeutils.extract_bits(pixal_quality, 6, end=9, new_name="b1_quality").eq(0)
        b3_qual = geeutils.extract_bits(pixal_quality, 10, end=13, new_name="b1_quality").eq(0)
        b4_qual = geeutils.extract_bits(pixal_quality, 14, end=17, new_name="b1_quality").eq(0)
        b6_qual = geeutils.extract_bits(pixal_quality, 22, end=25, new_name="b1_quality").eq(0)
        b7_qual = geeutils.extract_bits(pixal_quality, 26, end=29, new_name="b1_quality").eq(0)

        qual_mask = b1_qual.And(b2_qual).And(b3_qual).And(b4_qual).And(b6_qual).And(b7_qual)

        mask = env_mask.And(qual_mask)
        return geeutils.rescale(img).updateMask(mask)

class Landsat9(Dataset):
    def __init__(
        self,
        *args,
        asset_id="LANDSAT/LC09/C02/T1_L2",
        use_qa=True,
        **kwargs,
    ):
        """Initialize Landsat9 Dataset class
        Can theoretically be useds with any Landsat collection 2 surface reflectance collection (e.g. LANDSAT/LT05/C01/T1_SR)

        args:
            *args: positional arguments to pass to `Dataset` (i.e. `region`, `start_time`, `end_time`)
            asset_id (str): asset id of the Landsat earth engine collection. default="LANDSAT/LC08/C01/T1_SR"
            use_qa (bool, optional): boolean to determine to use a private `self.qa()` function. default=True
            rescale (bool, optional): boolean switch to convert units from scaled int (0-10000) to float (0-1). If false values will be scaled int. default = False
            **kwargs (optional): addtional arbitrary keywords to pass to `Dataset`
        """
        super(Landsat9, self).__init__(
            *args, asset_id=asset_id, use_qa=use_qa, **kwargs
        )

        self.collection = self.collection.select(
            self.BANDREMAP.get("landsat8"), self.BANDREMAP.get("new")
        )

        return

    @decorators.keep_attrs
    def qa(self, img):
        """Custom QA masking method for Landsat9 surface reflectance dataset"""
        qa_band = img.select("QA_PIXEL")
        qa_flag = int('111111',2)
        sat_mask = img.select('QA_RADSAT').eq(0);
        mask = qa_band.bitwiseAnd(qa_flag).eq(0).And(sat_mask)
        return geeutils.rescale(img, scale = 0.0000275, offset = -0.2).updateMask(mask)


class Landsat8(Dataset):
    def __init__(
        self,
        *args,
        asset_id="LANDSAT/LC08/C02/T1_L2",
        use_qa=True,
        **kwargs,
    ):
        """Initialize Landsat8 Dataset class
        Can theoretically be useds with any Landsat surface reflectance collection (e.g. LANDSAT/LT05/C01/T1_SR)

        args:
            *args: positional arguments to pass to `Dataset` (i.e. `region`, `start_time`, `end_time`)
            asset_id (str): asset id of the Landsat earth engine collection. default="LANDSAT/LC08/C01/T1_SR"
            use_qa (bool, optional): boolean to determine to use a private `self.qa()` function. default=True
            rescale (bool, optional): boolean switch to convert units from scaled int (0-10000) to float (0-1). If false values will be scaled int. default = False
            **kwargs (optional): addtional arbitrary keywords to pass to `Dataset`
        """
        super(Landsat8, self).__init__(
            *args, asset_id=asset_id, use_qa=use_qa, **kwargs
        )

        self.collection = self.collection.select(
            self.BANDREMAP.get("landsat8"), self.BANDREMAP.get("new")
        )

        return

    @decorators.keep_attrs
    def qa(self, img):
        """Custom QA masking method for Landsat8 surface reflectance dataset"""
        qa_band = img.select("QA_PIXEL")
        qa_flag = int('111111',2)
        sat_mask = img.select('QA_RADSAT').eq(0);
        mask = qa_band.bitwiseAnd(qa_flag).eq(0).And(sat_mask)
        return geeutils.rescale(img, scale = 0.0000275, offset = -0.2).updateMask(mask)


class Landsat7(Dataset):
    def __init__(
        self,
        *args,
        asset_id="LANDSAT/LE07/C02/T1_L2",
        use_qa=True,
        apply_band_adjustment=False,
        **kwargs,
    ):
        """Initialize Landsat7 Dataset class
        Can theoretically be useds with any Landsat surface reflectance collection (e.g. LANDSAT/LT05/C01/T1_SR)

        args:
            *args: positional arguments to pass to `Dataset` (i.e. `region`, `start_time`, `end_time`)
            asset_id (str): asset id of the Landsat7 earth engine collection. default="LANDSAT/LE07/C01/T1_SR"
            use_qa (bool, optional): boolean to determine to use a private `self.qa()` function. default=True
            apply_band_adjustment (bool, optional): boolean switch to apply linear band pass equation to convert values to Landsat8. default=False
            rescale (bool, optional): boolean switch to convert units from scaled int (0-10000) to float (0-1). If false values will be scaled int. default = False
            **kwargs (optional): addtional arbitrary keywords to pass to `Dataset`
        """
        super(Landsat7, self).__init__(
            *args, asset_id=asset_id, use_qa=use_qa, **kwargs
        )

        coll = self.collection.select(
            self.BANDREMAP.get("landsat7"), self.BANDREMAP.get("new")
        )

        if apply_band_adjustment:
            # band bass adjustment coefficients taken from Roy et al., 2016 http://dx.doi.org/10.1016/j.rse.2015.12.024
            # slope coefficients
            self.gain = ee.Image.constant(
                [0.8474, 0.8483, 0.9047, 0.8462, 0.8937, 0.9071]
            )
            # y-intercept coefficients
            self.bias = ee.Image.constant(
                [0.0003, 0.0088, 0.0061, 0.0412, 0.0254, 0.0172]
            ).multiply(10000)
            coll = coll.map(self.band_pass_adjustment)

        self.collection = coll

        return

    @decorators.keep_attrs
    def qa(self, img):
        """Custom QA masking method for Landsat7 surface reflectance dataset"""
        qa_band = img.select("QA_PIXEL")
        qa_flag = int('111111',2)
        sat_mask = img.select('QA_RADSAT').eq(0);
        mask = qa_band.bitwiseAnd(qa_flag).eq(0).And(sat_mask)
        return geeutils.rescale(img, scale = 0.0000275, offset = -0.2).updateMask(mask)


class Landsat5(Dataset):
    def __init__(
        self,
        *args,
        asset_id="LANDSAT/LT05/C02/T1_L2",
        use_qa=True,
        apply_band_adjustment=False,
        **kwargs,
    ):
        """Initialize Landsat7 Dataset class
        Can theoretically be useds with any Landsat surface reflectance collection (e.g. LANDSAT/LT05/C01/T1_SR)

        args:
            *args: positional arguments to pass to `Dataset` (i.e. `region`, `start_time`, `end_time`)
            asset_id (str): asset id of the Landsat7 earth engine collection. default="LANDSAT/LE07/C01/T1_SR"
            use_qa (bool, optional): boolean to determine to use a private `self.qa()` function. default=True
            apply_band_adjustment (bool, optional): boolean switch to apply linear band pass equation to convert values to Landsat8. default=False
            rescale (bool, optional): boolean switch to convert units from scaled int (0-10000) to float (0-1). If false values will be scaled int. default = False
            **kwargs (optional): addtional arbitrary keywords to pass to `Dataset`
        """
        super(Landsat5, self).__init__(
            *args, asset_id=asset_id, use_qa=use_qa, **kwargs
        )

        coll = self.collection.select(
            self.BANDREMAP.get("landsat7"), self.BANDREMAP.get("new")
        )

        if apply_band_adjustment:
            # TODO: get harmonization coefficients for TM to OLI
            # slope coefficients
            self.gain = ee.Image.constant(
                [
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                ]
            )
            # y-intercept coefficients
            self.bias = ee.Image.constant(
                [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                ]
            ).multiply(10000)
            coll = coll.map(self.band_pass_adjustment)

        self.collection = coll

        return

    @decorators.keep_attrs
    def qa(self, img):
        """Custom QA masking method for Landsat7 surface reflectance dataset"""
        qa_band = img.select("QA_PIXEL")
        qa_flag = int('111111',2)
        sat_mask = img.select('QA_RADSAT').eq(0);
        mask = qa_band.bitwiseAnd(qa_flag).eq(0).And(sat_mask)
        return geeutils.rescale(img, scale = 0.0000275, offset = -0.2).updateMask(mask)


class Sentinel2(Dataset):
    def __init__(
        self,
        *args,
        asset_id="COPERNICUS/S2_SR_HARMONIZED",
        use_qa=True,
        apply_band_adjustment=False,
        **kwargs,
    ):
        """Initialize Sentinel2 Dataset class

        args:
            *args: positional arguments to pass to `Dataset` (i.e. `region`, `start_time`, `end_time`)
            asset_id (str): asset id of the Sentinel2 earth engine collection. default="COPERNICUS/S2_SR"
            use_qa (bool, optional): boolean to determine to use a private `self.qa()` function. default=True
            apply_band_adjustment (bool, optional): boolean switch to apply linear band pass equation to convert values to Landsat8. default=False
            rescale (bool, optional): boolean switch to convert units from scaled int (0-10000) to float (0-1). If false values will be scaled int. default = False
            **kwargs (optional): addtional arbitrary keywords to pass to `Dataset`
        """
        super(Sentinel2, self).__init__(
            *args, asset_id=asset_id, use_qa=use_qa, **kwargs
        )

        coll = self.collection.select(
            self.BANDREMAP.get("sen2"), self.BANDREMAP.get("new")
        )

        if use_qa:
            coll = coll.filter(ee.Filter.eq('cloud_mask_success', 1))

        if apply_band_adjustment:
            # band bass adjustment coefficients taken HLS project https://hls.gsfc.nasa.gov/algorithms/bandpass-adjustment/
            # slope coefficients
            self.gain = ee.Image.constant(
                [0.9778, 1.0053, 0.9765, 0.9983, 0.9987, 1.003]
            )
            # y-intercept coefficients
            self.bias = ee.Image.constant(
                [-0.00411, -0.00093, 0.00094, -0.0001, -0.0015, -0.0012]
            ).multiply(10000)
            coll = coll.map(self.band_pass_adjustment)

        self.collection = coll

        return


    def deduplicate(self, inplace=False):
        """Deduplicate Sentinel2 images by removing images with the same data take identifier"""
        def _dedupe(id):
            """Helper function to remove duplicate images"""
            id_collection = (
                self.collection
                .filter(
                    ee.Filter.stringStartsWith("DATATAKE_IDENTIFIER", ee.String(id))
                )
                .sort('PROCESSING_BASELINE')
            )
            image = (
                id_collection
                .mosaic()
            )
  
            imgbounds = id_collection.map(lambda x: x.geometry(1e3))
  
            start_time = id_collection.aggregate_array('system:time_start').reduce(ee.Reducer.min())
            end_time = id_collection.aggregate_array('system:time_start').reduce(ee.Reducer.max())
  
            return image.clipToCollection(imgbounds).set(
                {
                    'system:time_start': start_time, 
                    'system:time_end':end_time,
                    'system:index':id
                }
            )
        
        datatakes = (
            self.collection
            .aggregate_histogram("DATATAKE_IDENTIFIER")
            .keys()
            .map(lambda x: ee.String(x).slice(0,-7))
        )

        images = ee.ImageCollection.fromImages(
            datatakes.map(_dedupe)
        )

        return self._inplace_wrapper(images, inplace)


    @decorators.keep_attrs
    def qa(self, img):
        """Custom QA masking method for Sentinel2 surface reflectance dataset"""
        CLD_PRB_THRESH = 40
        NIR_DRK_THRESH = 0.175 * 1e4
        CLD_PRJ_DIST = 3
        BUFFER = 100
        CRS = img.select(0).projection()

        # Get s2cloudless image, subset the probability band.
        cld_prb = ee.Image(
            ee.ImageCollection("COPERNICUS/S2_CLOUD_PROBABILITY")
            .filterDate(self.start_time, self.end_time)
            .filterBounds(self.region)
            .filter(ee.Filter.eq("system:index", img.get("system:index")))
            .first()
        ).select("probability")

        worked = cld_prb.bandNames().length().gt(0)

        # Condition s2cloudless by the probability threshold value.
        is_cloud = cld_prb.gt(CLD_PRB_THRESH)

        # Identify water pixels from the SCL band, invert.
        not_water = img.select("SCL").neq(6)

        # Identify dark NIR pixels that are not water (potential cloud shadow pixels).
        dark_pixels = img.select("B8").lt(NIR_DRK_THRESH).multiply(not_water)

        # Determine the direction to project cloud shadow from clouds (assumes UTM projection).
        shadow_azimuth = ee.Number(90).subtract(
            ee.Number(img.get("MEAN_SOLAR_AZIMUTH_ANGLE"))
        )

        # Project shadows from clouds for the distance specified by the CLD_PRJ_DIST input.
        cld_proj = (
            is_cloud.directionalDistanceTransform(shadow_azimuth, CLD_PRJ_DIST * 10)
            .reproject(**{"crs": CRS, "scale": 120})
            .select("distance")
            .mask()
        )

        # Identify the intersection of dark pixels with cloud shadow projection.
        is_shadow = cld_proj.multiply(dark_pixels)

        # Combine cloud and shadow mask, set cloud and shadow as value 1, else 0.
        is_cld_shdw = is_cloud.add(is_shadow).gt(0)

        # Remove small cloud-shadow patches and dilate remaining pixels by BUFFER input.
        # 20 m scale is for speed, and assumes clouds don't require 10 m precision.
        is_cld_shdw = (
            is_cld_shdw.focal_min(2)
            .focal_max(BUFFER * 2 / 20)
            .reproject(**{"crs": CRS, "scale": 60})
            .rename("cloudmask")
        )

        # Subset reflectance bands and update their masks, return the result.
        return geeutils.rescale(img).select("B.*").updateMask(is_cld_shdw.Not()).set({'cloud_mask_success':worked})



import ee
from ee.ee_exception import EEException
import math
import string
import random
import datetime
from hydrafloods import decorators
import hydrafloods.indices as all_indices


# helper function to convert qa bit image to flag
def extract_bits(image, start, end=None, new_name=None):
    """Function to conver qa bits to binary flag image

    args:
        image (ee.Image): qa image to extract bit from
        start (int): starting bit for flag
        end (int | None, optional): ending bit for flag, if None then will only use start bit. default = None
        new_name (str | None, optional): output name of resulting image, if None name will be {start}Bits. default = None

    returns:
        ee.Image: image with extract bits
    """

    newname = new_name if new_name is not None else f"{start}Bits"

    if (start == end) or (end is None):
        # perform a bit shift with bitwiseAnd
        return image.select([0], [newname]).bitwiseAnd(1 << start)
    else:
        # Compute the bits we need to extract.
        pattern = 0
        for i in range(start, end):
            pattern += int(math.pow(2, i))

        # Return a single band image of the extracted QA bits, giving the band
        # a new name.
        return image.select([0], [newname]).bitwiseAnd(pattern).rightShift(start)


def get_geoms(img):
    """Helper function to get geometry from image

    args:
        img (ee.Image): image to get geometry from

    returns:
        ee.Geometry: geometry of image
    """
    return img.geometry()


def export_image(
    image,
    region,
    asset_id=None,
    description=None,
    scale=1000,
    crs="EPSG:4326",
    pyramiding=None,
    export_type="toAsset",
    folder=None,
):
    """Function to wrap image export with EE Python API

    args:
        image (ee.Image): image to export
        region (ee.Geometry): region to export image
        asset_id (str | None, optional): asset ID to export image to\
            if None then asset_id will be a random string. default = None
        description (str | None, optional): description to identify image export/
            if None then description will be random string. default = None
        scale (int, optional): resolution in meters to export image to. default = 1000
        crs (str, optional): epsg code to export image to. default = "EPSG:4326"
        pyramiding (dict | None, optional): dictionary defining band pyramiding scheme.
            if None then "mean" will be used as default for all bands. default = None
        export_type (str, optional) : method by which to export the image.
            Default is 'toAsset', can also be 'toDrive'.
        folder (str | None, optional): target folder to export for 'toDrive'/
            if None then export should go to root of Drive

    """
    if (asset_id is None) or (type(asset_id) != str):
        asset_id = "".join(
            random.SystemRandom().choice(string.ascii_letters) for _ in range(8)
        ).lower()
    if (description is None) or (type(description) != str):
        description = "".join(
            random.SystemRandom().choice(string.ascii_letters) for _ in range(8)
        ).lower()
    if type(export_type) != str:
        raise TypeError(
            f"Input for export_type not a string, was a " f"{type(export_type)}."
        )
    elif (export_type != "toAsset") and (export_type != "toDrive"):
        raise ValueError(
            "Invalid input for export_type, must be " '"toAsset" or "toDrive".'
        )
    if (folder is not None) and (type(folder) != str):
        raise TypeError(f"Input for folder was not a string, was a " f"{type(folder)}")
    # get serializable geometry for export
    export_region = region.bounds(maxError=10).getInfo()["coordinates"]

    if pyramiding is None:
        pyramiding = {".default": "mean"}

    # set export process
    if export_type == "toAsset":
        export = ee.batch.Export.image.toAsset(
            image,
            description=description,
            assetId=asset_id,
            scale=scale,
            region=export_region,
            maxPixels=1e13,
            crs=crs,
            pyramidingPolicy=pyramiding,
        )
    elif export_type == "toDrive":
        export = ee.batch.Export.image.toDrive(
            image,
            description=description,
            folder=folder,
            scale=scale,
            region=export_region,
            maxPixels=1e13,
            crs=crs,
        )
    # start export process
    export.start()

    return


def batch_export(
    collection,
    collection_asset,
    region=None,
    prefix=None,
    suffix=None,
    scale=1000,
    crs="EPSG:4326",
    pyramiding=None,
    export_type="toAsset",
    folder=None,
    metadata=None,
    verbose=False,
):
    """Function to export each image in a collection
    Wraps `export_image` will set YYYYMMdd formatted time in file name

    args:
        collection (ee.ImageCollection): image collection to export
        collection_asset (str): image collection asset ID to export to
        region (ee.Geometry): region to export image
        prefix (str): prefix string to add before time info in name
        suffix (str): suffix string to add after time info in name
        scale (int, optional): resolution in meters to export image to. default = 1000
        crs (str, optional): epsg code to export image to. default = "EPSG:4326"
        pyramiding (dict | None, optional): dictionary defining band pyramiding scheme.
            if None then "mean" will be used as default for all bands. default = None
        export_type (str, optional) : method by which to export the image.
            Default is 'toAsset', can also be 'toDrive'.
        folder (str | None, optional): target folder to export for 'toDrive'/
                if None then export should go to root of Drive
        metadata (dict | None, optional):
        verbose (bool, optional):

    """
    if type(collection) is not ee.imagecollection.ImageCollection:
        try:
            collection = getattr(collection, "collection")
        except Exception as e:
            raise TypeError(
                "argument collection needs to be either of type ee.ImageCollection "
                "or hydrafloods.hfCollection"
            )

    n = collection.size()
    exportImages = collection.sort("system:time_start", False).toList(n)
    nIter = n.getInfo()

    for i in range(nIter):
        img = ee.Image(exportImages.get(i))
        if metadata is not None:
            img = img.set(metadata)

        t = img.get("system:time_start").getInfo()
        date = datetime.datetime.utcfromtimestamp(t / 1e3).strftime("%Y%m%d%H%M%S")

        if region is None:
            region = img.geometry()

        exportName = date
        if prefix is not None:
            exportName = f"{prefix}_" + exportName
        if suffix is not None:
            exportName = exportName + f"_{suffix}"

        description = exportName
        if verbose:
            print(f"running export for {description}")

        if not collection_asset.endswith("/"):
            collection_asset += "/"

        exportName = collection_asset + description

        export_image(
            img,
            region,
            asset_id=exportName,
            description=description,
            scale=scale,
            crs=crs,
            pyramiding=pyramiding,
            export_type=export_type,
            folder=folder,
        )

    return


@decorators.keep_names
@decorators.keep_attrs
def rescale(img, scale=0.0001, offset=0):
    """Function to linearly rescale units using user defined scale and offset

    args:
        img (ee.Image): image to rescale
        scale (float,optional): scale value (i.e. slope of linear equation). default = 0.0001
        offset (float, optional): offset value (i.e. y-intercept). default = 0

    returns:
        ee.Image: rescaled image
    """
    return img.multiply(scale).add(offset)


@decorators.keep_names
@decorators.keep_attrs
def power_to_db(img):
    """Function to convert SAR units from power to dB

    args:
        img (ee.Image): SAR power image to convert to dB

    returns:
        ee.Image: dB SAR image
    """
    return ee.Image(10).multiply(img.log10())


@decorators.keep_names
@decorators.keep_attrs
def db_to_power(img):
    """Function to convert SAR units from dB to power

    args:
        img (ee.Image): SAR dB image to convert to power

    returns:
        ee.Image: power SAR image
    """
    return ee.Image(10).pow(img.divide(10))


@decorators.keep_attrs
def add_indices(img, indices=["mndwi"]):
    """Function to calculate multiple band indices and add to image as bands

    args:
        img (ee.Image): image to calculate indices from
        indices (list[str], optional): list of strings of index names to calculate.
            can use any named index function in geeutils. default = ["mndwi"]

    returns:
        ee.Image: image object with added indices
    """

    # loop through each index and append to images list
    cat_bands = [img]
    for index in indices:
        index_func = getattr(all_indices, index)
        cat_bands.append(index_func(img))

    # return images as concatenated bands
    return ee.Image.cat(cat_bands)


def tile_region(
    region, grid_size=0.1, intersect_geom=None, contain_geom=None, centroid_within=None
):
    """Function to create a feature collection of tiles covering a region

    args:
        region (ee.Geometry): region to create tile grid over
        grid_size (float, optional): resolution in decimal degrees to create tiles. default = 0.1
        intersect_geom (ee.Geometry | None, optional): geometry object to filter tiles that intesect with
            geometry useful for filtering tiles that are created over oceans with no data. default = None
        contain_geom (ee.Geometry | None, optional): geometry object to filter tiles that are contained within
            geometry useful for filtering tiles that are only in an area. default = None

    returns:
        ee.FeatureCollection: collection of feature tiles at a given grid_size over a region
    """
    # nesting grid construction along y and then x coordinates
    def constuctGrid(i):
        """Closure function to contruct grid"""

        def contructXGrid(j):
            j = ee.Number(j)
            box = (
                ee.Geometry.Rectangle(
                    [j, i, j.add(grid_res), i.add(grid_res)],
                    "epsg:4326",
                    geodesic=False,
                )
            )
            if contain_geom is not None:
                keep = contain_geom.contains(box, maxError=500)

            elif intersect_geom is not None:
                keep = box.intersects(intersect_geom, maxError=500)

            elif centroid_within is not None:
                keep = box.centroid().intersects(centroid_within, maxError=500)

            return ee.Feature(box,{"ul_lat":i.add(grid_res), "ul_lon":j, "keep":keep} )

        i = ee.Number(i)
        out = ee.List.sequence(west, east.subtract(grid_res), grid_res).map(
            contructXGrid, True
        )
        return out

    if (contain_geom is not None) and (intersect_geom is not None):
        raise ValueError(
            "contains and intersection keywords are mutually exclusive, please define only one"
        )

    bounds = region.bounds(maxError=1000)
    coords = ee.List(bounds.coordinates().get(0))
    grid_res = ee.Number(grid_size)

    west = ee.Number(ee.List(coords.get(0)).get(0))
    south = ee.Number(ee.List(coords.get(0)).get(1))
    east = ee.Number(ee.List(coords.get(2)).get(0))
    north = ee.Number(ee.List(coords.get(2)).get(1))

    west = ee.Algorithms.If(
        west.lt(0),
        west.subtract(west.mod(grid_res).add(grid_res)),
        west.subtract(west.mod(grid_res)),
    )
    south = ee.Algorithms.If(
        south.lt(0),
        south.subtract(south.mod(grid_res).add(grid_res)),
        south.subtract(south.mod(grid_res)),
    )
    east = east.add(grid_res.subtract(east.mod(grid_res)))
    north = north.add(grid_res.subtract(north.mod(grid_res)))

    grid = ee.FeatureCollection(
        ee.List.sequence(south, north.subtract(grid_res), grid_res)
        .map(constuctGrid)
        .flatten()
    ).filter(ee.Filter.eq("keep",True))

    return grid


def country_bbox(country_name, max_error=1000):
    """Function to get a bounding box geometry of a country

    args:
        country_name (str): US-recognized country name
        max_error (float,optional): The maximum amount of error tolerated when
            performing any necessary reprojection. default = 100

    returns:
        ee.Geometry: geometry of country bounding box
    """

    all_countries = ee.FeatureCollection("USDOS/LSIB_SIMPLE/2017")
    return (
        all_countries.filter(ee.Filter.eq("country_na", country_name))
        .geometry(max_error)
        .bounds(max_error)
    )

def admin_bbox(admin_name, level=0, max_error=1000):
    """Function to get a bounding box geometry of an administrative area
    args:
        admin_name (str): US-recognized country name
        max_error (float,optional): The maximum amount of error tolerated when
            performing any necessary reprojection. default = 100

    returns:
        ee.Geometry: geometry of country bounding box
    """

    if level not in range(0,3):
        raise ValueError(f"Administrative level is 0-2, provided level {level} is outside of range")

    admin_bounds = ee.FeatureCollection("FAO/GAUL/2015/level2")
    return (
        admin_bounds.filter(ee.Filter.eq(f"ADM{level}_NAME", admin_name))
        .geometry(max_error)
        .bounds(max_error)
    )


import ee
import math
from hydrafloods import geeutils, decorators


@decorators.keep_attrs
def slope_correction(
    image,
    elevation,
    model="volume",
    buffer=0,
    scale=1000,
    in_units="db",
    out_units="same",
):
    """This function applies the slope correction on a Sentinel-1 image.
    Function based on https:# doi.org/10.3390/rs12111867.
    Adapted from https:# github.com/ESA-PhiLab/radiometric-slope-correction/blob/master/notebooks/1%20-%20Generate%20Data.ipynb

    args:
        image (ee.Image): Sentinel-1 to perform correction on
        elevation (ee.Image): Input DEM to calculate slope corrections from
        model (str, optional): physical reference model to be applied. Options are 'volume' or 'surface'.
            default = volume
        buffer (int, optional): buffer in meters for layover/shadow mask. If zero then no buffer will be applied. default = 0
        scale (int, optional): reduction scale to process satellite heading compared to ground. Increasing will reduce
            chance of OOM errors but reduce local scale correction accuracy. default = 1000

    returns:
        ee.Image: slope corrected SAR imagery with look and local incidence angle bands

    raises:
        NotImplementedError: when keyword model is not of 'volume' or 'surface'
    """

    def _volumetric_model_SCF(theta_iRad, alpha_rRad):
        """Closure funnction for calculation of volumetric model SCF

        args:
            theta_iRad (ee.Image): incidence angle in radians
            alpha_rRad (ee.Image): slope steepness in range

        returns:
            ee.Image
        """

        # model
        nominator = (ninetyRad.subtract(theta_iRad).add(alpha_rRad)).tan()
        denominator = (ninetyRad.subtract(theta_iRad)).tan()
        return nominator.divide(denominator)

    def _surface_model_SCF(theta_iRad, alpha_rRad, alpha_azRad):
        """Closure funnction for calculation of direct model SCF

        args:
            theta_iRad (ee.Image): incidence angle in radians
            alpha_rRad (ee.Image): slope steepness in range
            alpha_azRad (ee.Image): slope steepness in azimuth

        returns:
            ee.Image
        """

        # model
        nominator = (ninetyRad.subtract(theta_iRad)).cos()
        denominator = alpha_azRad.cos().multiply(
            (ninetyRad.subtract(theta_iRad).add(alpha_rRad)).cos()
        )

        return nominator.divide(denominator)

    def _erode(image, distance):
        """Closure function to buffer raster values

        args:
            image (ee.Image): image that should be buffered
            distance (int): distance of buffer in meters

        returns:
            ee.Image
        """

        d = (
            image.Not()
            .unmask(1)
            .fastDistanceTransform(10)
            .sqrt()
            .multiply(ee.Image.pixelArea().sqrt())
        )

        return image.updateMask(d.gt(distance))

    def _masking(alpha_rRad, theta_iRad, buffer):
        """Closure function for masking of layover and shadow

        args:
            alpha_rRad (ee.Image): slope steepness in range
            theta_iRad (ee.Image): incidence angle in radians
            buffer (int): buffer in meters

        returns:
            ee.Image
        """
        # layover, where slope > radar viewing angle
        layover = alpha_rRad.lt(theta_iRad).rename("layover")

        # shadow
        shadow = alpha_rRad.gt(
            ee.Image.constant(-1).multiply(ninetyRad.subtract(theta_iRad))
        ).rename("shadow")

        # add buffer to layover and shadow
        if buffer > 0:
            layover = _erode(layover, buffer)
            shadow = _erode(shadow, buffer)

        # combine layover and shadow
        no_data_mask = layover.And(shadow).rename("no_data_mask")

        return no_data_mask

    # get the image geometry and projection
    geom = image.geometry(scale)
    proj = image.select(1).projection()
    angle_band = image.select("angle")

    # image to convert angle to radians
    to_radians = ee.Image.constant((math.pi / 180))
    # create a 90 degree image in radians
    ninetyRad = ee.Image.constant(90).multiply(to_radians)

    # calculate the look direction
    heading = (
        ee.Terrain.aspect(image.select("angle"))
        .reduceRegion(ee.Reducer.mean(), geom, scale)
        .get("aspect")
    )

    if in_units not in ["db", "power"]:
        raise ValueError(
            "could not understand input units. needs to be either 'db' or 'power'"
        )

    if out_units not in ["same", "other", None]:
        raise ValueError(
            "could not understand output units option. needs to be either 'same' or 'other'"
        )

    # Sigma0 to Power of input image
    if in_units == "db":
        sigma0Pow = geeutils.db_to_power(image)
    else:
        sigma0Pow = image

    # the numbering follows the article chapters
    # 2.1.1 Radar geometry
    theta_iRad = image.select("angle").multiply(to_radians)
    phi_iRad = ee.Image.constant(heading).multiply(to_radians)

    # 2.1.2 Terrain geometry
    alpha_sRad = (
        ee.Terrain.slope(elevation)
        .select("slope")
        .multiply(to_radians)
        .setDefaultProjection(proj)
    )

    phi_sRad = (
        ee.Terrain.aspect(elevation)
        .select("aspect")
        .multiply(to_radians)
        .setDefaultProjection(proj)
    )

    # 2.1.3 Model geometry
    # reduce to 3 angle
    phi_rRad = phi_iRad.subtract(phi_sRad)

    # slope steepness in range (eq. 2)
    alpha_rRad = (alpha_sRad.tan().multiply(phi_rRad.cos())).atan()

    # slope steepness in azimuth (eq 3)
    alpha_azRad = (alpha_sRad.tan().multiply(phi_rRad.sin())).atan()

    # local incidence angle (eq. 4)
    theta_liaRad = (
        alpha_azRad.cos().multiply((theta_iRad.subtract(alpha_rRad)).cos())
    ).acos()
    theta_liaDeg = theta_liaRad.multiply(180 / math.pi)

    # 2.2
    # Gamma_nought
    gamma0 = sigma0Pow.divide(theta_iRad.cos())
    gamma0dB = geeutils.db_to_power(gamma0)

    if model == "volume":
        scf = _volumetric_model_SCF(theta_iRad, alpha_rRad)

    elif model == "surface":
        scf = _surface_model_SCF(theta_iRad, alpha_rRad, alpha_azRad)

    else:
        raise NotImplementedError(
            f"Defined model, {model}, has not been implemented. Options are 'volume' or 'surface'"
        )

    # apply model for Gamm0_f
    gamma0_flat = gamma0.divide(scf)

    if out_units == "same" and in_units == "db":
        gamma0_flatDB = geeutils.power_to_db(gamma0_flat).select(["^(V).*"])

    elif out_units != "same" and in_units == "power":
        gamma0_flatDB = geeutils.power_to_db(gamma0_flat).select(["^(V).*"])

    else:
        gamma0_flatDB = gamma0_flat.select(["^(V).*"])

    # calculate layover and shadow mask
    masks = _masking(alpha_rRad, theta_iRad, buffer)

    return (
        gamma0_flatDB.updateMask(masks)
        .addBands(angle_band)
        .addBands(theta_liaDeg.rename("local_inc_angle"))
    )


@decorators.keep_attrs
def illumination_correction(image, elevation, model="rotation", scale=90, sensor="LC8"):
    """This function applies a terrain correction to optical imagery based on solar and viewing geometry

    args:
        image (ee.Image): Optical image to perform correction on
        elevation (ee.Image): Input DEM to calculate illumination corrections from
        model (str, optional): correction model to be applied. Options are 'cosine', 'c', 'scsc', or 'rotation'
            default = rotation
        scale (int, optional): reduction scale to process satellite heading compared to ground. Increasing will reduce
            chance of OOM errors but reduce local scale correction accuracy. default = 90
        sensor (str, optional): name of sensor to correct. options are 'LC8' or 'S2' (lower case also accepted).
            default = LC8

    returns:
        ee.Image: illumination corrected optical imagery

    raises:
        NotImplementedError: when keyword sensor is not of 'LC8' or 'S2'
        NotImplementedError: when keyword model is not of 'cosine', 'c', 'scsc', or 'rotation'
    """

    def _get_band_coeffs(band_name):
        """Closure function to find illumination correction fit across the different bands

        args:
            band_name (str | ee.String): band name to find correction coefficients for
        """

        #  Create the image to apply the linear regression.The first band
        # is the cosi and the second band is the response variable, the reflectance (the bands).
        # L (y) = a + b*cosi(x); a = intercept, b = slope
        #  Dependent: Reflectance
        y = image.select([band_name])
        #  create an image with the three variables by concatenating them
        reg_image = ee.Image.cat([cosi, one, y])
        #  specify the linear regression reducer
        lr_reducer = ee.Reducer.linearRegression(numX=2, numY=1)
        #  fit the model
        fit = reg_image.reduceRegion(
            reducer=lr_reducer, geometry=image.geometry(), scale=scale, maxPixels=1e10
        )

        #  Get the coefficients as a nested list, cast it to an array, and get
        #  just the selected column
        slope = ee.Array(fit.get("coefficients")).get([0, 0])
        intercept = ee.Array(fit.get("coefficients")).get([1, 0])

        return ee.List([slope, intercept])

    if sensor.lower() == "lc8":
        sz_property = "SOLAR_ZENITH_ANGLE"
        sa_property = "SOLAR_AZIMUTH_ANGLE"
    elif sensor.lower() == "s2":
        sz_property = "MEAN_SOLAR_ZENITH_ANGLE"
        sa_property = "MEAN_SOLAR_AZIMUTH_ANGLE"
    else:
        raise NotImplementedError(
            f"Selected sensor, {sensor}, is not available. Options are 'LC8' or 'S2' (lower case also accepted)"
        )

    # value convert angle to radians
    to_radians = ee.Number((math.pi / 180))
    # constand image of 1
    one = ee.Image.constant(1).rename("one")

    # calculate terrain info from elevation data
    terrain = ee.Algorithms.Terrain(elevation)
    # Extract slope in radians for each pixel in the image
    p = terrain.select(["slope"]).multiply(to_radians)
    # Extract aspect in radians for each pixel in the image
    o = terrain.select(["aspect"]).multiply(to_radians)
    # Extract solar zenith angle from the image
    z = ee.Image.constant(ee.Number(image.get(sz_property)).multiply(to_radians))
    # Extract solar azimuth from the image
    az = ee.Image.constant(ee.Number(image.get(sa_property)).multiply(to_radians))

    cosao = (o.subtract(az)).cos()
    # cos(ϕa−ϕo)
    # Calculate the cosine of the local solar incidence for every pixel in the image in radians (cosi=cosp*cosz+sinp*sinz*cos(ϕa−ϕo)
    cosi = image.expression(
        "((cosp * cosz) + (sinp * sinz * cosao))",
        {
            "cosp": p.cos(),
            "cosz": z.cos(),
            "sinp": p.sin(),
            "sinz": z.sin(),
            "cosao": cosao,
        },
    )

    if model == "cosine":
        # if cosine model correction, return early as we don't need to do extra processing
        return image.expression(
            "((image * cosz) / cosi) ", {"image": image, "cosz": z.cos(), "cosi": cosi}
        )

    bnames = image.bandNames()
    ab = ee.Array(bnames.map(_get_band_coeffs))

    # get the coefficients as images
    a = ee.Image(ee.Array(ab.slice(1, 0, 1))).arrayProject([0]).arrayFlatten([bnames])
    b = ee.Image(ee.Array(ab.slice(1, 1, 2))).arrayProject([0]).arrayFlatten([bnames])
    C = b.divide(a)

    if model == "c":
        newimage = image.expression(
            "((image * (cosz + C)) / (cosi + C))",
            {"image": image, "cosz": z.cos(), "cosi": cosi, "C": C},
        )

    elif model == "scsc":
        newimage = image.expression(
            "((image * ((cosp * cosz) + C))/(cosi + C))",
            {"image": image, "cosp": p.cos(), "cosz": z.cos(), "cosi": cosi, "C": C},
        )

    elif model == "rotation":
        # Apply the empirical rotation model
        newimage = image.expression(
            "image - a * (cosi - cosz)",
            {"image": image, "cosz": z.cos(), "cosi": cosi, "a": a},
        )

    else:
        raise NotImplementedError(
            f"Defined model, {model}, has not been implemented. Options are 'cosine', 'c', 'scsc', or 'rotation'"
        )

    return newimage



import ee
from ee.ee_exception import EEException
import math
import datetime
from hydrafloods import decorators


@decorators.keep_attrs
def fuzzy_gaussian(img, midpoint, spread):
    """Fuzzy membership function through a Gaussian or normal distribution based around a
    user-specified midpoint (which is assigned a membership of 1) with a defined spread.
    https://desktop.arcgis.com/en/arcmap/10.3/analyze/arcpy-spatial-analyst/fuzzygaussian-class.htm

    args:
        img (ee.Image): The input image whose values will be scaled from 0 to 1
        midpoint (float): user-defined value with a fuzzy membership of 1
        spread (float): the spread of the Gaussian function. The spread generally
            ranges from 0.01 to 1, with the larger the value results in a steeper
            distribution around the midpoint

    returns:
        ee.Image: output floating-point image with values ranging from 0 to 1
    """
    bandNames = img.bandNames()

    spread = ee.Image.constant(spread)
    midpoint = ee.Image.constant(midpoint)
    e = ee.Image.constant(math.e)

    gauss = e.expression(
        "e ** (-s * (x-m)**2)", {"e": e, "x": img, "s": spread, "m": midpoint}
    )

    return gauss.rename(bandNames)


@decorators.keep_attrs
def fuzzy_near(img, midpoint, spread):
    """Fuzzy membership function around a specific value which is defined by a user-defined midpoint
     (which is assigned a membership of 1), with a defined spread decreasing to zero.
    https://desktop.arcgis.com/en/arcmap/10.3/analyze/arcpy-spatial-analyst/fuzzylarge-class.htm

    args:
        img (ee.Image): The input image whose values will be scaled from 0 to 1
        midpoint (float): user-defined value with a fuzzy membership of 1
        spread (float): the spread of the Near function. The spread generally
            ranges from 0.001 to 1, with the larger the value results in a steeper
            distribution from the midpoint

    returns:
        ee.Image: output floating-point image with values ranging from 0 to 1
    """
    bandNames = img.bandNames()

    spread = ee.Image.constant(spread)
    midpoint = ee.Image.constant(midpoint)
    one = ee.Image.constant(1)

    near = one.expression(
        "o / (o + s * (x - m)**2)", {"o": one, "x": img, "s": spread, "m": midpoint}
    )

    return near.rename(bandNames)


@decorators.keep_attrs
def fuzzy_large(img, midpoint, spread):
    """Fuzzy membership function where the larger input values have membership closer to 1.
    The function is defined by a user-specified midpoint (which is assigned a membership of 0.5)
    with a defined spread.
    https://desktop.arcgis.com/en/arcmap/10.3/analyze/arcpy-spatial-analyst/fuzzylarge-class.htm

    args:
        img (ee.Image): The input image whose values will be scaled from 0 to 1
        midpoint (float): user-defined value with a fuzzy membership of 0.5
        spread (float): the spread of the Large function. The spread generally
            ranges from 1 to 10, with the larger the value results in a steeper
            distribution from the midpoint

    returns:
        ee.Image: output floating-point image with values ranging from 0 to 1
    """
    bandNames = img.bandNames()

    spread = ee.Image(spread)
    midpoint = ee.Image(midpoint)
    one = ee.Image.constant(1)

    large = one.expression(
        "o / ( o + (x / m) ** (s * -1))",
        {"o": one, "x": img, "s": spread, "m": midpoint},
    )

    return large.rename(bandNames)


@decorators.keep_attrs
def fuzzy_mslarge(img, mean_scaling, std_scaling, region=None, scale=90):
    """Fuzzy membership through a function based on the mean and standard deviation,
    with the larger values having a membership closer to 1.
    https://desktop.arcgis.com/en/arcmap/10.3/analyze/arcpy-spatial-analyst/fuzzymslarge-class.htm

    args:
        img (ee.Image): The input image whose values will be scaled from 0 to 1
        mean_scaling (float): multiplier for the mean of the input values in the MSLarge function equation.
        std_scaling (float): multiplier for the standard deviation of the input values in the MSLarge function equation.
        region (ee.Geometry | None, optional): region to calculate mean/stdDev image statistics,
            if set to None will use img.geometry(). default = None
        scale (int, optional): scale at which to perform reduction operations, setting higher will prevent OOM errors. default = 90

    returns:
        ee.Image: output floating-point image with values ranging from 0 to 1
    """
    if region is None:
        region = img.geometry()

    bandNames = img.bandNames()

    mean_scale = ee.Image.constant(mean_scaling)
    std_scale = ee.Image.constant(std_scaling)
    one = ee.Image.constant(1)

    reducer = ee.Reducer.mean().combine(ee.Reducer.stdDev(), None, True)

    stats = ee.Dictionary(
        img.reduceRegion(reducer, region, scale, bestEffort=True, tileScale=16)
    ).toImage()

    mslarge = one.expression(
        "o - (b * s ) / (x - (a * m) + (b * s))",
        {
            "o": one,
            "x": img,
            "a": stats.select(".*(mean)$"),
            "b": stats.select(".*(stdDev)$"),
            "m": mean_scale,
            "s": std_scale,
        },
    )

    mask = img.gte(stats.select(".*(mean)$").multiply(mean_scale))

    return mslarge.multiply(mask).rename(bandNames)


@decorators.keep_attrs
def fuzzy_small(img, midpoint, spread):
    """Fuzzy membership function with the smaller input values having membership closer to 1.
    The function is defined by a user-specified midpoint (which is assigned a membership of 0.5)
    with a defined spread.
    https://desktop.arcgis.com/en/arcmap/10.3/analyze/arcpy-spatial-analyst/fuzzysmall-class.htm

    args:
        img (ee.Image): The input image whose values will be scaled from 0 to 1
        midpoint (float): user-defined value with a fuzzy membership of 0.5
        spread (float): the spread of the Small function. The spread generally
            ranges from 1 to 10, with the larger the value results in a steeper
            distribution from the midpoint

    returns:
        ee.Image: output floating-point image with values ranging from 0 to 1
    """
    bandNames = img.bandNames()

    spread = ee.Image(spread)
    midpoint = ee.Image(midpoint)
    one = ee.Image.constant(1)

    small = one.expression(
        "o / (o + (x / m) ** s)", {"o": one, "x": img, "s": spread, "m": midpoint}
    )

    return small.rename(bandNames)


@decorators.keep_attrs
def fuzzy_mssmall(img, mean_scale, std_scale, region=None, scale=90):
    """Fuzzy membership through a function based on the mean and standard deviation,
    with the smaller values having a membership closer to 1.
    https://desktop.arcgis.com/en/arcmap/10.3/analyze/arcpy-spatial-analyst/fuzzymssmall-class.htm

    args:
        img (ee.Image): The input image whose values will be scaled from 0 to 1
        mean_scaling (float): multiplier for the mean of the input values in the MSSmall function equation.
        std_scaling (float): multiplier for the standard deviation of the input values in the MSSmall function equation.
        region (ee.Geometry | None, optional): region to calculate mean/stdDev image statistics,
            if set to None will use img.geometry(). default = None
        scale (int, optional): scale at which to perform reduction operations, setting higher will prevent OOM errors. default = 90

    returns:
        ee.Image: output floating-point image with values ranging from 0 to 1
    """
    if region is None:
        region = img.geometry()

    bandNames = img.bandNames()

    mean_scale = ee.Image.constant(mean_scale)
    std_scale = ee.Image.constant(std_scale)
    one = ee.Image.constant(1)

    reducer = ee.Reducer.mean().combine(ee.Reducer.stdDev(), None, True)

    stats = ee.Dictionary(
        img.reduceRegion(reducer, region, scale, bestEffort=True, tileScale=16)
    ).toImage()

    mssmall = one.expression(
        "(b * s ) / (x - (a * m) + (b * s))",
        {
            "o": one,
            "x": img,
            "a": stats.select(".*(mean)$"),
            "b": stats.select(".*(stdDev)$"),
            "m": mean_scale,
            "s": std_scale,
        },
    )

    mask = img.gte(stats.select(".*(mean)$").multiply(mean_scale))

    return mssmall.multiply(mask).rename(bandNames)


@decorators.keep_attrs
def fuzzy_linear(img, min_val, max_val):
    """Fuzzy membership function through a linear transformation between the user-specified minimum
    value, a membership of 0, to the user-defined maximum value, which is assigned a membership of 1.
    If the minimum value is less than the maximum, the linear function will have a positive slope.
    If the minimum value is greater than the maximum, the linear function will have a negative slope.
    https://desktop.arcgis.com/en/arcmap/10.3/analyze/arcpy-spatial-analyst/fuzzylinear-class.htm

    args:
        img (ee.Image): The input image whose values will be scaled linearly from 0 to 1
        minimum (float): The value that will have a membership of 0
        maximum (float): The value that will have a membership of 1

    returns:
        ee.Image: output floating-point image with values ranging from 0 to 1
    """
    bandNames = img.bandNames()

    if ~isinstance(min_val, ee.Number):
        min_val = ee.Number(min_val)

    if ~isinstance(max_val, ee.Number):
        max_val = ee.Number(max_val)

    invert = min_val.gt(max_val)

    minimum = ee.Algorithms.If(invert, max_val, min_val)
    maximum = ee.Algorithms.If(invert, min_val, max_val)

    linear = img.unitScale(minimum, maximum)

    linear = ee.Algorithms.If(invert, ee.Image.constant(1).subtract(linear), linear)
    linear = ee.Image(linear).clamp(0, 1)

    return linear.rename(bandNames)


def fuzzy_zmf(img, min_val, max_val):

    """
    Z-function fuzzy membership generator.
    """
    bandNames = img.bandNames()

    if ~isinstance(min_val, ee.Number):
        min_val = ee.Number(min_val)

    if ~isinstance(max_val, ee.Number):
        max_val = ee.Number(max_val)

    invert = min_val.gt(max_val)

    a = ee.Image.constant(ee.Algorithms.If(invert, max_val, min_val))
    b = ee.Image.constant(ee.Algorithms.If(invert, min_val, max_val))

    zmf = ee.Image(1)

    m1 = a.lte(img).And(img.lt(a.add(b).divide(2)))
    y1 = img.expression(
        "1 - 2 * ((img - a) / (b - a)) ** 2", {"img": img, "a": a, "b": b}
    )

    m2 = a.add(b).divide(2).lte(img).And(img.lte(b))
    y2 = img.expression("2 * ((img - b) / (b - a)) ** 2", {"img": img, "a": a, "b": b})

    m3 = img.gte(b)

    zmf = zmf.where(m1, y1).where(m2, y2).where(m3, ee.Image(0))

    zmf = ee.Image(ee.Algorithms.If(invert, ee.Image(1).subtract(zmf), zmf))

    return zmf.rename(bandNames)


def fuzzy_or(img_list):
    """Fuzzy Or overlay returning the maximum value of the input images.
    This technique is useful when you want to identify the highest membership values
    for any of the input criteria

    args:
        img_list (list[ee.Image]): list of ee.Image objects to overlay together

    returns:
        ee.Image: output floating-point image with overlayed values
    """
    in_img = ee.Image.cat([img_list])

    out_img = in_img.reduce(ee.Reducer.max()).rename("fuzzy_or")

    return out_img


def fuzzy_and(img_list):
    """Fuzzy And overlay returning the minimum value of the input images.
    This technique is useful when you want to identify the least common denominator
    for the membership of all the input criteria

    args:
        img_list (list[ee.Image]): list of ee.Image objects to overlay together

    returns:
        ee.Image: output floating-point image with overlayed values
    """
    in_img = ee.Image.cat([img_list])

    out_img = in_img.reduce(ee.Reducer.min()).rename("fuzzy_and")

    return out_img


def fuzzy_product(img_list):
    """Fuzzy Product overlay which multiplies each of the fuzzy values togetherfor all the input images.
    The resulting product will be less than any of the input, and when a member of many sets is input,
    the value can be very small.

    args:
        img_list (list[ee.Image]): list of ee.Image objects to overlay together

    returns:
        ee.Image: output floating-point image with overlayed values
    """
    in_img = ee.Image.cat([img_list])

    out_img = in_img.reduce(ee.Reducer.product()).rename("fuzzy_product")

    return out_img


def fuzzy_sum(img_list):
    """Fuzzy Product overlay add the fuzzy values for all the input images.
    The resulting sum is an increasing linear combination function that is based on the number of
    criteria entered into the analysis

    args:
        img_list (list[ee.Image]): list of ee.Image objects to overlay together

    returns:
        ee.Image: output floating-point image with overlayed values
    """
    one = ee.Image.constant(1)

    in_img = one.subtract(ee.Image.cat([img_list]))

    out_img = one.subtract(in_img.reduce(ee.Reducer.product())).rename("fuzzy_sum")

    return out_img


def fuzzy_gamma(img_list, gamma=0.5):
    """Fuzzy Gamma is an algebraic product of fuzzy Product and fuzzy Sum,
    which are both raised to the power of gamma. If the specified gamma is 1,
    the output is the same as fuzzy Sum; if gamma is 0, the output is the same as fuzzy Product

    args:
        img_list (list[ee.Image]): list of ee.Image objects to overlay together
        gamma (float): the gamma value to be used when weighting Product vs Sum

    returns:
        ee.Image: output floating-point image with overlayed values
    """
    gamma = ee.Image.constant(gamma)
    one = ee.Image.constant(1)
    a = fuzzy_sum(img_list)
    b = fuzzy_product(img_list)

    out_img = a.pow(gamma).multiply(b.pow(one.subtract(gamma))).rename("fuzzy_gamma")

    return out_img


def fuzzy_weighted(img_list, weights):
    """Overlays several rasters, multiplying each by their given weight and summing them together.

    args:
        img_list (list[ee.Image]): list of ee.Image objects to overlay together
        weights (list[float]): list of weight values to assign images, must be in order

    returns:
        ee.Image: output floating-point image with overlayed values
    """
    weights = ee.Image.constant(weights)

    in_img = ee.Image.cat([img_list])

    out_img = in_img.multiply(weights).reduce(ee.Reducer.sum()).rename("fuzzy_weighted")

    return out_img



from hydrafloods.datasets import *
from hydrafloods.geeutils import *
from hydrafloods.thresholding import *
from hydrafloods.filtering import *
from hydrafloods.corrections import *
from hydrafloods.indices import *
from hydrafloods.fuzzy import *
from hydrafloods.depths import *
from hydrafloods.floods import *
from hydrafloods import fetch, utils

from hydrafloods import fusion

# from hydrafloods import *

__version__ = "2023.10.14"



# coding: utf-8 VIIRS DNNS

import ee
import numpy as np
import math
import matplotlib.pyplot as pl


def perm_water_mask():
    return ee.Image("MODIS/MOD44W/MOD44W_005_2000_02_24").select(["water_mask"], ["b1"])


def DEM():
    return ee.Image("CGIAR/SRTM90_V4")


# ALOS DEM
# ee.Image('JAXA/ALOS/AW3D30_V1_1')
# SRTM DEM
# ee.Image('USGS/SRTMGL1_003')


# def GEE_classifier(img, name_classifier, arguments={}):
#     arguments = {
#         "training_image": perm_water_mask(),
#         "training_band": "b1",
#         "training_region": img,
#     }
#     c_arguments = {
#         "subsampling": 0.1,
#         "max_classification": 2,
#         "classifier_mode": "classification",
#         "name_classifier": name_classifier,
#     }
#     arguments.update(c_arguments)
#     classes1 = img.trainClassifier(**arguments)
#     classes2 = classify(classes1).select(["classification"], ["b1"])
#     return classes2


def dnns(img):

    kernel_scale = 40
    VIIRS_ave_scale = 500

    k1 = ee.Kernel.square(kernel_scale, "pixels", False)

    composite = img["I1"].addBands(img["I2"]).addBands(img["I3"])

    # classes = GEE_classifier(img, "Pegasos", {"classifier_mode": "probability"})
    classes = ee.Image([0.5, 0.5])
    water = classes.gte(0.9)
    land = classes.lte(0.1)
    mix = water.Not().And(land.Not())
    ave_water = water.mask(water).multiply(composite)
    ave_water_img = ee.Image([ave_water["I1"], ave_water["I2"], ave_water["I3"]])

    N_nmin_water = 1
    N_water = water.convolve(k1)

    water_rf = (
        water.multiply(composite)
        .convolve(k1)
        .multiply(N_water.gte(N_nmin_water))
        .divide(N_water)
    )
    water_rf = water_rf.add(ave_water_img.multiply(water_rf.Not()))

    ave_land = composite.mask(land)

    ave_land_img = ee.Image([ave_land["I1"], ave_land["I2"], ave_land["I3"]])

    # Constraints
    R1 = img["I1"].divide(img["I3"])
    R2 = img["I2"].divide(img["I3"])
    R3 = R1.subtract(water_rf.select("I1").divide(img["I3"]))
    R4 = R2.subtract(water_rf.select("I2").divide(img["I3"]))

    NR1 = R1.neighborhoodToBands(k1)
    NR2 = R2.neighborhoodToBands(k1)
    NI1 = img["I1"].neighborhoodToBands(k1)
    NI2 = img["I2"].neighborhoodToBands(k1)
    NI3 = img["I3"].neighborhoodToBands(k1)

    M1 = (NR1.gt(R3)).And(NR1.lt(R1))
    M2 = (NR2.gt(R4)).And(NR2.lt(R2))
    nLP = M1.And(M2)

    NnLP = nLP.reduce(ee.Reducer.sum())
    ave_nI1 = NI1.multiply(nLP).reduce(ee.Reducer.sum()).divide(NnLP)
    ave_nI2 = NI2.multiply(nLP).reduce(ee.Reducer.sum()).divide(NnLP)
    ave_nI3 = NI3.multiply(nLP).reduce(ee.Reducer.sum()).divide(NnLP)

    N_nmin_land = 1
    ave_land = ave_nI1.addBands(ave_nI2).addBands(ave_nI3)
    ave_land = ave_land.multiply(NnLP.gte(N_nmin_land)).add(
        ave_land_img.multiply(NnLP.lt(N_nmin_land))
    )

    ave_landI3 = ave_land.select("I3")
    f_water = (
        (ave_landI3.subtract(img["I3"]))
        .divide(ave_landI3.subtract(water_rf.select("I3")))
        .clamp(0, 1)
    )
    f_water = f_water.add(water).subtract(land).clamp(0, 1)

    return f_water


def DEM_downscale(img, f_water):

    VIIRS_Pixel_meters = 500
    DEM_d = img.DEM()

    water_present = f_water.gt(0.0)

    h_min = DEM_d.mask(f_water).focal_min(VIIRS_Pixel_meters, "square", "meters")
    h_max = DEM_d.mask(f_water).focal_max(VIIRS_Pixel_meters, "square", "meters")

    water_high = h_min.add(h_max.subtract(h_min).multiply(f_water))
    water_high = water_high.multiply(f_water.lt(1.0)).multiply(f_water.gt(0.0))
    return f_water.eq(1.0).select(["elevation"], ["I1"])



import setuptools
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="hydrafloods",
    version="2023.10.14",
    description="HYDrologic Remote sensing Analysis for Floods",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/servir-mekong/hydra-floods",
    packages=setuptools.find_packages(),
    author="Kel Markert",
    author_email="kel.markert@gmail.com",
    license="GNU GPL v3.0",
    zip_safe=False,
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "hydrafloods = hydrafloods.hfcli:main",
        ]
    },
    install_requires=["simplecmr", "earthengine-api", "gcsfs", "fire", "pipetools"],
)



import ee
import math
from hydrafloods import geeutils, decorators


@decorators.keep_names
@decorators.keep_attrs
def lee_sigma(img, window=9, sigma=0.9, looks=4, tk=7, keep_bands=["angle"]):
    """Lee Sigma speckle filtering algorithm.
    Implemented from interpreting https://doi.org/10.1109/TGRS.2008.2002881

    args:
        img (ee.Image): Earth engine image object. Expects that imagery is a SAR image
        window (int, optional): moving window size to apply filter (i.e. a value of 9 == 9x9 window). default = 9
        sigma (float, optional): sigma lookup value from table 1 in paper. default = 0.9
        looks (int, optional): look intensity value from table 1 in paper. default = 4
        tk (int, optional): threshold value to determine values in window as point targets. default = 7
        keep_bands (list[str], optional): list of band names to drop during filtering and include in the result
            default = ["angle"]

    returns:
        ee.Image: filtered SAR image using the Lee Sigma algorithm
    """
    band_names = img.bandNames()
    if keep_bands is not None:
        keep_img = img.select(keep_bands)
        proc_bands = band_names.removeAll(keep_bands)
    else:
        proc_bands = band_names

    img = img.select(proc_bands)

    midPt = (window // 2) + 1 if (window % 2) != 0 else window // 2
    kernelWeights = ee.List.repeat(ee.List.repeat(1, window), window)
    kernel = ee.Kernel.fixed(window, window, kernelWeights, midPt, midPt)

    targetWeights = ee.List.repeat(ee.List.repeat(1, 3), 3)
    targetkernel = ee.Kernel.fixed(3, 3, targetWeights, 1, 1)

    # Lookup table for range and eta values for intensity
    sigmaLookup = ee.Dictionary(
        {
            1: ee.Dictionary(
                {
                    0.5: ee.Dictionary({"A1": 0.436, "A2": 1.92, "η": 0.4057}),
                    0.6: ee.Dictionary({"A1": 0.343, "A2": 2.21, "η": 0.4954}),
                    0.7: ee.Dictionary({"A1": 0.254, "A2": 2.582, "η": 0.5911}),
                    0.8: ee.Dictionary({"A1": 0.168, "A2": 3.094, "η": 0.6966}),
                    0.9: ee.Dictionary({"A1": 0.084, "A2": 3.941, "η": 0.8191}),
                    0.95: ee.Dictionary({"A1": 0.043, "A2": 4.840, "η": 0.8599}),
                }
            ),
            2: ee.Dictionary(
                {
                    0.5: ee.Dictionary({"A1": 0.582, "A2": 1.584, "η": 0.2763}),
                    0.6: ee.Dictionary({"A1": 0.501, "A2": 1.755, "η": 0.3388}),
                    0.7: ee.Dictionary({"A1": 0.418, "A2": 1.972, "η": 0.4062}),
                    0.8: ee.Dictionary({"A1": 0.327, "A2": 2.260, "η": 0.4819}),
                    0.9: ee.Dictionary({"A1": 0.221, "A2": 2.744, "η": 0.5699}),
                    0.95: ee.Dictionary({"A1": 0.152, "A2": 3.206, "η": 0.6254}),
                }
            ),
            3: ee.Dictionary(
                {
                    0.5: ee.Dictionary({"A1": 0.652, "A2": 1.458, "η": 0.2222}),
                    0.6: ee.Dictionary({"A1": 0.580, "A2": 1.586, "η": 0.2736}),
                    0.7: ee.Dictionary({"A1": 0.505, "A2": 1.751, "η": 0.3280}),
                    0.8: ee.Dictionary({"A1": 0.419, "A2": 1.865, "η": 0.3892}),
                    0.9: ee.Dictionary({"A1": 0.313, "A2": 2.320, "η": 0.4624}),
                    0.95: ee.Dictionary({"A1": 0.238, "A2": 2.656, "η": 0.5084}),
                }
            ),
            4: ee.Dictionary(
                {
                    0.5: ee.Dictionary({"A1": 0.694, "A2": 1.385, "η": 0.1921}),
                    0.6: ee.Dictionary({"A1": 0.630, "A2": 1.495, "η": 0.2348}),
                    0.7: ee.Dictionary({"A1": 0.560, "A2": 1.627, "η": 0.2825}),
                    0.8: ee.Dictionary({"A1": 0.480, "A2": 1.804, "η": 0.3354}),
                    0.9: ee.Dictionary({"A1": 0.378, "A2": 2.094, "η": 0.3991}),
                    0.95: ee.Dictionary({"A1": 0.302, "A2": 2.360, "η": 0.4391}),
                }
            ),
        }
    )

    # extract data from lookup
    looksDict = ee.Dictionary(sigmaLookup.get(ee.String(str(looks))))
    sigmaImage = ee.Dictionary(looksDict.get(ee.String(str(sigma)))).toImage()
    a1 = sigmaImage.select("A1")
    a2 = sigmaImage.select("A2")
    aRange = a2.subtract(a1)
    eta = sigmaImage.select("η").pow(2)

    img = geeutils.db_to_power(img)

    # MMSE estimator
    mmseMask = img.gte(a1).Or(img.lte(a2))
    mmseIn = img.updateMask(mmseMask)
    oneImg = ee.Image(1)
    z = mmseIn.reduceNeighborhood(ee.Reducer.mean(), kernel, None, True)
    varz = mmseIn.reduceNeighborhood(ee.Reducer.variance(), kernel)
    varx = (varz.subtract(z.abs().pow(2).multiply(eta))).divide(oneImg.add(eta))
    b = varx.divide(varz)
    mmse = oneImg.subtract(b).multiply(z.abs()).add(b.multiply(mmseIn))

    # workflow
    z99 = ee.Dictionary(
        img.reduceRegion(
            reducer=ee.Reducer.percentile([99], None, 255, 0.001, 1e6),
            geometry=img.geometry(),
            scale=10,
            bestEffort=True,
        )
    ).toImage()

    overThresh = img.gte(z99)

    K = overThresh.reduceNeighborhood(ee.Reducer.sum(), targetkernel, None, True)

    retainPixel = K.gte(tk)
    xHat = geeutils.power_to_db(img.updateMask(retainPixel).unmask(mmse))

    output = ee.Image(xHat).rename(proc_bands)

    if keep_bands is not None:
        output = output.addBands(keep_img)

    return output


# The RL speckle filter
@decorators.keep_names
@decorators.keep_attrs
def refined_lee(image, keep_bands=["angle"]):
    """Refined Lee speckle filtering algorithm.
    Algorithm adapted from https://groups.google.com/g/google-earth-engine-developers/c/ExepnAmP-hQ/m/7e5DnjXXAQAJ

    args:
        image (ee.Image): Earth engine image object. Expects that imagery is a SAR image

    returns:
        ee.Image: filtered SAR image using the Refined Lee algorithm
    """
    # TODO: include keep bands...maybe one-shot filtering if using keep_bands???
    def apply_filter(b):
        """Closure function to apply the refined lee algorithm on individual bands"""
        b = ee.String(b)
        img = power.select(b)

        # img must be in natural units, i.e. not in dB!
        # Set up 3x3 kernels
        weights3 = ee.List.repeat(ee.List.repeat(1, 3), 3)
        kernel3 = ee.Kernel.fixed(3, 3, weights3, 1, 1, False)

        mean3 = img.reduceNeighborhood(ee.Reducer.mean(), kernel3)
        variance3 = img.reduceNeighborhood(ee.Reducer.variance(), kernel3)

        # Use a sample of the 3x3 windows inside a 7x7 windows to determine gradients and directions
        sample_weights = ee.List(
            [
                [0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 1, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 1, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 1, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 0],
            ]
        )

        sample_kernel = ee.Kernel.fixed(7, 7, sample_weights, 3, 3, False)

        # Calculate mean and variance for the sampled windows and store as 9 bands
        sample_mean = mean3.neighborhoodToBands(sample_kernel)
        sample_var = variance3.neighborhoodToBands(sample_kernel)

        # Determine the 4 gradients for the sampled windows
        gradients = sample_mean.select(1).subtract(sample_mean.select(7)).abs()
        gradients = gradients.addBands(
            sample_mean.select(6).subtract(sample_mean.select(2)).abs()
        )
        gradients = gradients.addBands(
            sample_mean.select(3).subtract(sample_mean.select(5)).abs()
        )
        gradients = gradients.addBands(
            sample_mean.select(0).subtract(sample_mean.select(8)).abs()
        )

        # And find the maximum gradient amongst gradient bands
        max_gradient = gradients.reduce(ee.Reducer.max())

        # Create a mask for band pixels that are the maximum gradient
        gradmask = gradients.eq(max_gradient)

        # duplicate gradmask bands: each gradient represents 2 directions
        gradmask = gradmask.addBands(gradmask)

        # Determine the 8 directions
        directions = (
            sample_mean.select(1)
            .subtract(sample_mean.select(4))
            .gt(sample_mean.select(4).subtract(sample_mean.select(7)))
            .multiply(1)
        )
        directions = directions.addBands(
            sample_mean.select(6)
            .subtract(sample_mean.select(4))
            .gt(sample_mean.select(4).subtract(sample_mean.select(2)))
            .multiply(2)
        )
        directions = directions.addBands(
            sample_mean.select(3)
            .subtract(sample_mean.select(4))
            .gt(sample_mean.select(4).subtract(sample_mean.select(5)))
            .multiply(3)
        )
        directions = directions.addBands(
            sample_mean.select(0)
            .subtract(sample_mean.select(4))
            .gt(sample_mean.select(4).subtract(sample_mean.select(8)))
            .multiply(4)
        )
        # The next 4 are the not() of the previous 4
        directions = directions.addBands(directions.select(0).Not().multiply(5))
        directions = directions.addBands(directions.select(1).Not().multiply(6))
        directions = directions.addBands(directions.select(2).Not().multiply(7))
        directions = directions.addBands(directions.select(3).Not().multiply(8))

        # Mask all values that are not 1-8
        directions = directions.updateMask(gradmask)

        # "collapse" the stack into a singe band image (due to masking, each pixel has just one value (1-8) in it's directional band, and is otherwise masked)
        directions = directions.reduce(ee.Reducer.sum())

        sample_stats = sample_var.divide(sample_mean.multiply(sample_mean))

        # Calculate localNoiseVariance
        sigmaV = (
            sample_stats.toArray()
            .arraySort()
            .arraySlice(0, 0, 5)
            .arrayReduce(ee.Reducer.mean(), [0])
        )

        # Set up the 7*7 kernels for directional statistics
        rect_weights = ee.List.repeat(ee.List.repeat(0, 7), 3).cat(
            ee.List.repeat(ee.List.repeat(1, 7), 4)
        )

        diag_weights = ee.List(
            [
                [1, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0, 0],
                [1, 1, 1, 0, 0, 0, 0],
                [1, 1, 1, 1, 0, 0, 0],
                [1, 1, 1, 1, 1, 0, 0],
                [1, 1, 1, 1, 1, 1, 0],
                [1, 1, 1, 1, 1, 1, 1],
            ]
        )

        rect_kernel = ee.Kernel.fixed(7, 7, rect_weights, 3, 3, False)
        diag_kernel = ee.Kernel.fixed(7, 7, diag_weights, 3, 3, False)

        # Create stacks for mean and variance using the original kernels. Mask with relevant direction.
        dir_mean = img.reduceNeighborhood(ee.Reducer.mean(), rect_kernel).updateMask(
            directions.eq(1)
        )
        dir_var = img.reduceNeighborhood(ee.Reducer.variance(), rect_kernel).updateMask(
            directions.eq(1)
        )

        dir_mean = dir_mean.addBands(
            img.reduceNeighborhood(ee.Reducer.mean(), diag_kernel).updateMask(
                directions.eq(2)
            )
        )
        dir_var = dir_var.addBands(
            img.reduceNeighborhood(ee.Reducer.variance(), diag_kernel).updateMask(
                directions.eq(2)
            )
        )

        # and add the bands for rotated kernels
        for i in range(1, 4):
            dir_mean = dir_mean.addBands(
                img.reduceNeighborhood(
                    ee.Reducer.mean(), rect_kernel.rotate(i)
                ).updateMask(directions.eq(2 * i + 1))
            )
            dir_var = dir_var.addBands(
                img.reduceNeighborhood(
                    ee.Reducer.variance(), rect_kernel.rotate(i)
                ).updateMask(directions.eq(2 * i + 1))
            )
            dir_mean = dir_mean.addBands(
                img.reduceNeighborhood(
                    ee.Reducer.mean(), diag_kernel.rotate(i)
                ).updateMask(directions.eq(2 * i + 2))
            )
            dir_var = dir_var.addBands(
                img.reduceNeighborhood(
                    ee.Reducer.variance(), diag_kernel.rotate(i)
                ).updateMask(directions.eq(2 * i + 2))
            )

        # "collapse" the stack into a single band image (due to masking, each pixel has just one value in it's directional band, and is otherwise masked)
        dir_mean = dir_mean.reduce(ee.Reducer.sum())
        dir_var = dir_var.reduce(ee.Reducer.sum())

        # A finally generate the filtered value
        varX = dir_var.subtract(dir_mean.multiply(dir_mean).multiply(sigmaV)).divide(
            sigmaV.add(1.0)
        )

        b = varX.divide(dir_var)

        # return multi-band image band from array
        return (
            dir_mean.add(b.multiply(img.subtract(dir_mean)))
            .arrayProject([0])
            .arrayFlatten([["sum"]])
            .float()
        )

    band_names = image.bandNames()
    if keep_bands is not None:
        keep_img = image.select(keep_bands)
        proc_bands = band_names.removeAll(keep_bands)
    else:
        proc_bands = band_names

    image = image.select(proc_bands)

    power = geeutils.db_to_power(image)

    result = ee.ImageCollection(proc_bands.map(apply_filter)).toBands()

    output = geeutils.power_to_db(ee.Image(result)).rename(proc_bands)

    if keep_bands is not None:
        output = output.addBands(keep_img)

    return output


@decorators.keep_names
@decorators.keep_attrs
def gamma_map(img, window=7, enl=4.9, keep_bands=["angle"]):
    """Gamma Map speckle filtering algorithm.
    Algorithm adapted from https://groups.google.com/g/google-earth-engine-developers/c/a9W0Nlrhoq0/m/tnGMC45jAgAJ.

    args:
        img (ee.Image): Earth engine image object. Expects that imagery is a SAR image
        window (int, optional): moving window size to apply filter (i.e. a value of 7 == 7x7 window). default = 7
        enl (float, optional): equivalent number of looks (enl) per pixel from a SAR scan.
            See https://sentinel.esa.int/web/sentinel/user-guides/sentinel-1-sar/resolutions/level-1-ground-range-detected.
            default = 4.9
        keep_bands (list[str], optional): list of band names to drop during filtering and include in the result
            default = ["angle"]

    returns:
        ee.Image: filtered SAR image using the Gamma Map algorithm
    """

    band_names = img.bandNames()
    if keep_bands is not None:
        keep_img = img.select(keep_bands)
        proc_bands = band_names.removeAll(keep_bands)
    else:
        proc_bands = band_names

    img = img.select(proc_bands)

    # Square kernel, window should be odd (typically 3, 5 or 7)
    weights = ee.List.repeat(ee.List.repeat(1, window), window)
    midPt = (window // 2) + 1 if (window % 2) != 0 else window // 2

    # ~~(window/2) does integer division in JavaScript
    kernel = ee.Kernel.fixed(window, window, weights, midPt, midPt, False)

    # Convert image from dB to natural values
    nat_img = geeutils.db_to_power(img)

    # Get mean and variance
    mean = nat_img.reduceNeighborhood(ee.Reducer.mean(), kernel)
    variance = nat_img.reduceNeighborhood(ee.Reducer.variance(), kernel)

    # "Pure speckle" threshold
    ci = variance.sqrt().divide(mean)  # square root of inverse of enl

    # If ci <= cu, the kernel lies in a "pure speckle" area -> return simple mean
    cu = 1.0 / math.sqrt(enl)

    # If cu < ci < cmax the kernel lies in the low textured speckle area -> return the filtered value
    cmax = math.sqrt(2.0) * cu

    alpha = ee.Image(1.0 + cu * cu).divide(ci.multiply(ci).subtract(cu * cu))
    b = alpha.subtract(enl + 1.0)
    d = (
        mean.multiply(mean)
        .multiply(b)
        .multiply(b)
        .add(alpha.multiply(mean).multiply(nat_img).multiply(4.0 * enl))
    )
    f = b.multiply(mean).add(d.sqrt()).divide(alpha.multiply(2.0))

    caster = ee.Dictionary.fromLists(
        proc_bands, ee.List.repeat("float", proc_bands.length())
    )
    img1 = (
        geeutils.power_to_db(mean.updateMask(ci.lte(cu)))
        .rename(proc_bands)
        .cast(caster)
    )
    img2 = (
        geeutils.power_to_db(f.updateMask(ci.gt(cu)).updateMask(ci.lt(cmax)))
        .rename(proc_bands)
        .cast(caster)
    )
    img3 = img.updateMask(ci.gte(cmax)).rename(proc_bands).cast(caster)

    # If ci > cmax do not filter at all (i.e. we don't do anything, other then masking)
    output = (
        ee.ImageCollection([img1, img2, img3])
        .reduce(ee.Reducer.firstNonNull())
        .rename(proc_bands)
        .clip(img.geometry())
    )

    if keep_bands is not None:
        output = output.addBands(keep_img)

    # Compose a 3 band image with the mean filtered "pure speckle", the "low textured" filtered and the unfiltered portions
    return output


@decorators.keep_names
@decorators.keep_attrs
def p_median(img, window=5, keep_bands=["angle"]):
    """P-Median filter for smoothing imagery.
    Calculates the average from the median along cross and diagnal pixels of a window

    args:
        img (ee.Image): Earth engine image object to filter
        window (int, optional): moving window size to apply filter (i.e. a value of 5 == 5x5 window). default = 5

    returns:
        ee.Image: filtered image
    """

    def _band_filter(bname):
        selector = ee.List([bname])
        band_img = img.select(selector)
        hv_median = band_img.reduceNeighborhood(ee.Reducer.median(), hv_kernel)
        diag_median = band_img.reduceNeighborhood(ee.Reducer.median(), diag_kernel)
        return ee.Image(ee.Image.cat([hv_median, diag_median]).reduce("mean")).rename(
            selector
        )

    if window % 2 == 0:
        window += 1

    center_idx = (window - 1) // 2

    hv = [
        [1 if i == center_idx or j == center_idx else 0 for j in range(window)]
        for i in range(window)
    ]
    diag = [
        [1 if i == j or i == ((window - 1) - j) else 0 for j in range(window)]
        for i in range(window)
    ]

    # method based on ???
    band_names = img.bandNames()
    hv_weights = ee.List(hv)
    diag_weights = ee.List(diag)

    hv_kernel = ee.Kernel.fixed(window, window, hv_weights)
    diag_kernel = ee.Kernel.fixed(window, window, diag_weights)

    reduced_bands = ee.ImageCollection.fromImages(
        band_names.map(_band_filter)
    ).toBands()

    return reduced_bands.rename(band_names)


@decorators.keep_names
@decorators.keep_attrs
def perona_malik(img, n_iters=10, K=3, method=1):
    """Perona-Malik (anisotropic diffusion) convolution
    Developed by Gennadii Donchyts see https://groups.google.com/g/google-earth-engine-developers/c/umGlt5qIN1I/m/PD8lsJ7qBAAJ
    I(n+1, i, j) = I(n, i, j) + lambda * (cN * dN(I) + cS * dS(I) + cE * dE(I), cW * dW(I))

    args:
        img (ee.Image): Earth engine image object. Expects that imagery is a SAR image
        n_iters (int, optional): Number of interations to apply filter
            K (int,optional): moving window size to apply filter (i.e. a value of 7 == 7x7 window). default = 3
        method (int, optional): choose method 1 (default) or 2
    returns:
        ee.Image: filtered SAR image using the perona malik algorithm
    """

    def _method_1(dI_W, dI_E, dI_N, dI_S):
        cW = dI_W.pow(2).multiply(k1).exp()
        cE = dI_E.pow(2).multiply(k1).exp()
        cN = dI_N.pow(2).multiply(k1).exp()
        cS = dI_S.pow(2).multiply(k1).exp()

        return cW, cE, cN, cS

    def _method_2(dI_W, dI_E, dI_N, dI_S):
        cW = one.divide(one.add(dI_W.pow(2).divide(k2)))
        cE = one.divide(one.add(dI_E.pow(2).divide(k2)))
        cN = one.divide(one.add(dI_N.pow(2).divide(k2)))
        cS = one.divide(one.add(dI_S.pow(2).divide(k2)))

        return cW, cE, cN, cS

    # covnert db to natural units  before applying filter
    power = geeutils.db_to_power(img)

    dxW = ee.Kernel.fixed(3, 3, [[0, 0, 0], [1, -1, 0], [0, 0, 0]])
    dxE = ee.Kernel.fixed(3, 3, [[0, 0, 0], [0, -1, 1], [0, 0, 0]])
    dyN = ee.Kernel.fixed(3, 3, [[0, 1, 0], [0, -1, 0], [0, 0, 0]])
    dyS = ee.Kernel.fixed(3, 3, [[0, 0, 0], [0, -1, 0], [0, 1, 0]])

    one = ee.Image.constant(1.0)
    l = ee.Image.constant(0.2)
    k = ee.Image.constant(K)
    k1 = ee.Image.constant(-1.0 / K)
    k2 = k.pow(2)

    if method == 1:
        _method = _method_1
    elif method == 2:
        _method = _method_2
    else:
        raise NotImplementedError(
            "Could not determine algorithm to apply filter...options for `method` are 1 or 2"
        )

    for i in range(n_iters):
        dI_W = power.convolve(dxW)
        dI_E = power.convolve(dxE)
        dI_N = power.convolve(dyN)
        dI_S = power.convolve(dyS)

        cW, cE, cN, cS = _method(dI_W, dI_E, dI_N, dI_S)

        power = power.add(
            l.multiply(
                cN.multiply(dI_N)
                .add(cS.multiply(dI_S))
                .add(cE.multiply(dI_E))
                .add(cW.multiply(dI_W))
            )
        )

    # covnert natural to db units after filter is done
    img = geeutils.power_to_db(power)

    return img


@decorators.keep_names
@decorators.keep_attrs
def modified_median_zscore(img, fill_img=None):
    """Outlier detection and filling on complete DEM using the modified z-score and a median filter
    Method from  Iglewicz, B. and Hoaglin, D.C., 1993. How to detect and handle outliers (Vol. 16). Asq Press.

    args:
        img (ee.Image): Earth engine image object to apply filter on
        fill_img (ee.Image): Earth engine image object to fill values resulting from filter.
            If None provided, the fill operation will be on the input image. Default = None

    returns:
        ee.Image: filtered image
    """

    kernel = ee.Kernel.fixed(3, 3, [[1, 1, 1], [1, 1, 1], [1, 1, 1]])
    kernel_weighted = ee.Kernel.fixed(3, 3, [[1, 1, 1], [1, 0, 1], [1, 1, 1]])

    median = img.focal_median(kernel=kernel)
    median_weighted = img.focal_median(kernel=kernel_weighted)

    diff = img.subtract(median)

    mzscore = diff.multiply(0.6745).divide(diff.abs().focal_median(kernel=kernel))

    if fill_img:
        filled = fill_img.where(mzscore.gt(3.5), median_weighted)
    else:
        filled = img.where(mzscore.gt(3.5), median_weighted)

    return filled


@decorators.keep_names
@decorators.keep_attrs
def open_binary(img, window=3, neighborhood=None):
    """Opening morphological filter. Opening is the dilation of the erosion of
    values greater than 1.

    args:
        img (ee.Image): binary image to apply opening filter on
        window (int | ee.Number, optional): distance in pixels to consider for opening process
        neighborhood (int | ee.Number, optional): size of the neighborhood to perform fast
            distance trasnform. Smaller values speed up operations however must be greater
            than window. If no nerighborhood is provided, it will be double the window size.
            default = None

    returns:
        ee.Image: the opened binary image

    """
    if not isinstance(window, ee.Number):
        window = ee.Number(window)

    if neighborhood is None:
        neighborhood = window.multiply(2)

    erosion = img.Not().fastDistanceTransform(neighborhood).sqrt().gt(window)
    opened = erosion.fastDistanceTransform(neighborhood).sqrt().lt(window)
    return opened.updateMask(img.mask())


@decorators.keep_names
@decorators.keep_attrs
def close_binary(img, window=3, neighborhood=None):
    """Closing morphological filter. Closing is the erosion of the dialiation of
    values greater than 1.

    args:
        img (ee.Image): binary image to apply closing filter on
        window (int | ee.Number, optional): distance in pixels to consider for closing process
        neighborhood (int | ee.Number, optional): size of the neighborhood to perform fast
            distance trasnform. Smaller values speed up operations however must be greater
            than window. If no neighborhood is provided, it will be double the window size.
            default = None

    returns:
        ee.Image: the closed binary image

    """
    if not isinstance(window, ee.Number):
        window = ee.Number(window)

    if neighborhood is None:
        neighborhood = window.multiply(2)

    dialation = img.fastDistanceTransform(neighborhood).sqrt().lt(window)
    closed = dialation.Not().fastDistanceTransform(neighborhood).sqrt().gt(window)

    return closed.updateMask(img.mask())
